"""pywatemsedem grass strips processing functions"""
import logging
import tempfile
from pathlib import Path
from typing import Callable

import geopandas as gpd
import numpy as np
from scipy import signal

from pywatemsedem.defaults import PREFIX_TEMP
from pywatemsedem.geo.utils import (
    clean_up_tempfiles,
    estimate_width_of_polygon,
    saga_intersection,
    vct_to_rst_field,
)

# Add new kte scaling functions here
logger = logging.getLogger(__name__)


def _check_grass_strip_width(arr_width: np.array):
    """Check a number of conditions for the definition of grass strip widths array

    1. Grass strip widths (m) are rounded to their integer value.
    2. Grass strip widths (m) smaller than one are not allowed.

    Parameters
    ----------
    arr_width: numpy.ndarray
        Vector array with grass strips widths (m).

    Returns
    -------
    arr_width: numpy.ndarray
        Vector array with rounded grass strips widths (m).
    """

    if not np.issubdtype(arr_width.dtype, np.number):
        msg = (
            f"Grass strip widths array should be a numerical array, not"
            f" {arr_width.dtype}. Abort"
        )
        raise ValueError(msg)

    if np.any(np.isnan(arr_width)):
        msg = "Non defined grass strips widths (nan) are not allowed. Abort"
        raise ValueError(msg)

    # round to integer (m)
    arr_width = np.round(arr_width).astype(int)

    if np.any(arr_width < 1):
        arr_width[arr_width < 1] = 1
        msg = "Grass widths lower than one are observed, setting equal to one."
        logging.warning(msg)
    return arr_width


def scale_cfactor_linear(
    arr_width: np.ndarray,
    resolution: (int, float) = 20,
    upper_cfactor=0.37,
    lower_cfactor=0.01,
):
    """Scale function for ktc based on a lineair interpolation between the defined lower
    and upper boundary condtion, and the resolution.

    Parameters
    ----------
    arr_width: numpy.ndarray
        See :func:`pywatemsedem.grasstrips.scale_cfactor_with_grass_strip_width`
    resolution: int
        Spatial resolution of raster grid on which grass strips are projected.
    upper_cfactor: float, default 0.37
        Upper allowed C-factor.
    upper_cfactor: float, default 0.01
        Lower allowed C-factor.

    Returns
    -------
    arr_cfactor: numpy.ndarray
        See :func:`pywatemsedem.grasstrips.scale_cfactor_with_grass_strip_width`

    Notes
    -----
    The C-factor (:math:`C`) is weighted by the grass strips width (:math:`w`)
    and resolution (:math:`r`):

    .. math::

        C = 0.01*\\frac{w}{r}+0.37*\\frac{r-w}{r}

    Considering an lower and upper C-factor of 0.01 and 0.37, widths are capped to
    model resolution.

    References
    ----------
    Deproost, P., Renders, D., Van de Wauw, J., Van Ransbeeck, N., Verstraeten, G.,
    2018. Herkalibratie van WaTEM/SEDEM met het DHMV-II als hoogtemodel: eindrapport.
    Brussel.
    """
    arr_width = np.where(
        arr_width >= resolution,
        resolution,
        arr_width,
    )
    arr_cfactor = (lower_cfactor * (arr_width / resolution)) + (
        upper_cfactor * ((resolution - arr_width) / resolution)
    )

    return arr_cfactor


def scale_ktc_linear(
    arr_width: np.ndarray,
    resolution: (int, float) = 20,
    ktc_low: (int, float) = 3,
    ktc_high: (int, float) = 12,
):
    """Scale function for ktc based on a lineair interpolation between the defined lower
    and upper boundary condtion, and the resolution.

    Parameters
    ----------
    arr_width: numpy.ndarray
        See :func:`pywatemsedem.grasstrips.scale_ktc_with_grass_strip_width`
    resolution: int
        Spatial resolution of raster grid on which grass strips are projected.
    ktc_low: float
        The lower boundary to which scale the ktc value.
    ktc_high: float
        The upper boundary to which scale the ktc value.

    Returns
    -------
    arr_ktc: numpy.ndarray
        See :func:`pywatemsedem.grasstrips.scale_ktc_with_grass_strip_width`
    arr_ste: numpy.ndarray
        See :func:`pywatemsedem.grasstrips.scale_ktc_with_grass_strip_width`

    Notes
    -----
    1. ktc values for grass strips are determined by a lower and upper boundary of ktc.
    The weighting is defined by the width of the grass strip compared to the resolution
    of the model. This approach is defined in Deproost et al. (2018).

    2. The estimated STE values are determined by using the formula of Verstraete et
    al. (2006):

    .. math::

        STE (percent) = [1 -  * ktc_{var}/ktc_{high}]*100

    with:

        :math:`ktc_{var}` (1/m): the ktc values for the grass strips;

    References
    ----------
    Deproost, P., Renders, D., Van de Wauw, J., Van Ransbeeck, N., Verstraeten, G.,
    2018. Herkalibratie van WaTEM/SEDEM met het DHMV-II als hoogtemodel: eindrapport.
    Brussel.

    Verstraeten, G., Poesen, J., Gillijns, K., Govers, G., 2006. The use of riparian
    vegetated filter strips to reduce river sediment loads: an overestimated control
    measure? Hydrol. Process. 20, 4259–4267. https://doi.org/10.1002/hyp.6155

    """
    if np.any(arr_width > resolution):
        msg = (
            f"Grass strips larger than the resolution {resolution} are not allowed "
            f"for linear scaling of ktc value grass strips, setting to {resolution} m."
        )
        arr_width[arr_width > resolution] = resolution
        logger.warning(msg)

    w1 = arr_width / resolution
    w2 = 1 - arr_width / resolution
    arr_ktc = ktc_low * w1 + ktc_high * w2
    arr_ste = 1 - arr_ktc / ktc_high

    return arr_ktc, arr_ste


def scale_ktc_zhang(
    arr_width: np.ndarray,
    ktc_high: (int, float) = 12,
    k: (int, float) = 90.9,
    b: (int, float) = 0.446,
):
    """Scale function for ktc based on a Zhang et al. (2010).

    Parameters
    ----------
    arr_width: numpy.ndarray
        See :func:`pywatemsedem.grasstrips.scale_ktc_with_grass_strip_width`
    ktc_high: float
        The upper boundary to which scale the ktc value.
    k: float
        Maximum sediment trappping efficiency (K in Zhang et al., 2010)
    b: float
        Slope coefficient (see Zhang et al., 2010).

    Returns
    -------
    arr_ktc: numpy.ndarray
        See :func:`pywatemsedem.grasstrips.scale_ktc_with_grass_strip_width`
    arr_ste: numpy.ndarray
        See :func:`pywatemsedem.grasstrips.scale_ktc_with_grass_strip_width`

    Notes
    -----
    1. The ktc valeus for grass strips are determined by using emperical sediment
    trapping efficiency (STE) values in the equation of Verstraete et al. (2006):

    .. math::

        ktc_{var} = (ktc_{high}/100)*(1-STE/100)


    with:

        :math:`ktc_{var}` (1/m): the ktc values for the grass strips;

    The STE can be determined by using the emperical findings of Zhang et al. (2010):

    .. math::s


        STE (percent) = K * (1-e^{-b*w_{gs}})


    with:

        :math:`K` = 90.9;

        :math:`b` = 0.446;

        :math:`gs` (m) = width grass strips

    References
    ----------
    Zhang, X., Liu, X., Zhang, M., Dahlgren, R.A., Eitzel, M., 2010. A Review of
    Vegetated Buffers and a Meta-analysis of Their Mitigation Efficacy in Reducing
    Nonpoint Source Pollution. J. Environ. Qual. 39, 76–84.
    https://doi.org/10.2134/jeq2008.0496
    """
    arr_ste = k * (1 - np.exp(-b * arr_width)) / 100  # in decimals
    arr_ktc = (ktc_high) * (1 - arr_ste)

    return arr_ktc, arr_ste


# Add new kte scaling functions here
KTC_SCALING_FUNCTIONS = [scale_ktc_linear, scale_ktc_zhang]
CFACTOR_SCALING_FUNCTIONS = [scale_cfactor_linear]


def scale_ktc_with_grass_strip_width(
    arr_width: np.ndarray, scaling_function: Callable, **parameters
):
    """Scale ktc value for grass stripts according to the width of the grass strip.

    The ktc parameters (which determines the amount of sediment which is routed
    downstream) for a grass strip varies as a function of the width of the grass strip.

    Parameters
    ----------
    arr_width: numpy.ndarray
        Vector array with grass strips widths (m)
    scaling_function: callable
        Scaling function
    parameters:
        Scaling function parameters as required by the ``scaling function``.

    Returns
    -------
    arr_ktc: numpy.ndarray
        Vector array with ktc values for grass strips (1/m)
    arr_ste: numpy.ndarray
        Vector array of sediment trapping efficiency of grass strips
        (under the assumption of a flat plane, see Notes)

    Notes
    -----
    The theoretical 'sediment trapping efficiency' (STE) is interpreted as the
    trapping efficiency of a grass strips on a homogenous lineair slope. In other
    words, the slope plane can be defined as a flat plane (as in: can be defined in
    by two dimensions). This is also sometimes referred to as a linear slope (see
    Verstraete et al. (2006))

    References
    ----------
    Verstraeten, G., Poesen, J., Gillijns, K., Govers, G., 2006. The use of riparian
    vegetated filter strips to reduce river sediment loads: an overestimated control
    measure? Hydrol. Process. 20, 4259–4267. https://doi.org/10.1002/hyp.6155
    """
    if scaling_function not in KTC_SCALING_FUNCTIONS:
        msg = f"Ktc scale function {scaling_function} not implemented in pywatemsedem."
        raise IOError(msg)

    arr_width = _check_grass_strip_width(arr_width)

    arr_ktc, arr_ste = scaling_function(arr_width, **parameters)

    return arr_ktc, arr_ste


def estimate_ste(simulated_sediin, simulated_sediout):
    """Compute the estimated :math:`\\hat{STE}` based on simulated sediment output/input

    This function computes the sediment trapping efficiency based on the
    model-simulated incoming and outgoing sediment for each grass strip.

    Parameters
    ----------
    simulated_sediin: numpy.ndarray
        Simulated incoming sediment per grass strip
    simulated_sediout: numpy.ndarray
        Simulated outgoing sediment per grass strip

    Returns
    -------
    float
        Estimated STE (STEe)

    Notes
    -----
    1. The estimated :math:`\\hat{STE}`  is referred to with a hat to make a distinction
       between the theoretical STE defined in
       :func:`pywatemsedem.grasstrips.scale_ktc_with_grass_strip_width`
    2. Note that the STE can be negative in case there is more erosie produced in the
       grass strip then there is incoming into the grass strip.
    """
    return (1 - simulated_sediout / simulated_sediin) * 100


def scale_cfactor_with_grass_strips_width(
    arr_width: np.ndarray, scaling_function_cfactor: Callable, **parameters
):
    """Scale C-factor value for grass strips according to the width of the grass strip.

    Parameters
    ----------
    arr_width: numpy.ndarray
        Vector array with grass strips widths (m)
    scaling_function_cfactor: callable
        Scaling function
    parameters:
        Scaling function parameters as required by the ``scaling function``.

    Returns
    -------
    arr_cfactor: numpy.ndarray
        Vector array with C-factor values for grass strips (-)
    """
    if scaling_function_cfactor not in CFACTOR_SCALING_FUNCTIONS:
        msg = (
            f"Ktc scale function {scaling_function_cfactor} not implemented in "
            f"pywatemsedem."
        )
        raise IOError(msg)

    arr_width = _check_grass_strip_width(arr_width)

    arr_cfactor = scaling_function_cfactor(arr_width, **parameters)

    return arr_cfactor


def get_width_grass_strips(
    arr_width_gras_strips, arr_gras_polygon_perimeter, arr_gras_polygon_area, resolution
):
    """Get the width of the gras strips for gras strips with width being equal
    to np.nan 'unknown' or zero.

    If width cannot be estimated, estimate is based on polygon shape.

    Parameters
    ----------
    arr_width_gras_strips: numpy.ndarray or pandas.Series
        1D array holding in each row width of the gras strip.
        note: this can all be zeros or np.nan
    arr_gras_polygon_perimeter: numpy.ndarray or pandas.Series
        1D array holding in each row the perimeter of each gras polygon
    arr_gras_polygon_area: numpy.ndarray or pandas.Series
        1D array holding in each row the area of each gras polygon
    resolution: int

    Returns
    -------
    arr_width_gras_strips: numpy.ndarray or pandas.Series
        1D array holding in each row (estimated) width of the gras strip
    """
    arr_est_polygon_width = estimate_width_of_polygon(
        arr_gras_polygon_perimeter, arr_gras_polygon_area, nan_value=resolution
    )
    # fill in zeros with estimates
    arr_width_gras_strips = np.where(
        arr_width_gras_strips == 0,
        arr_est_polygon_width,
        arr_width_gras_strips,
    )
    # fill in np.nan with estimates
    arr_width_gras_strips = np.where(
        np.isnan(arr_width_gras_strips), arr_est_polygon_width, arr_width_gras_strips
    )

    return arr_width_gras_strips


def expand_grass_strips_with_triggers(
    arr_grass_strips: np.ndarray,
    arr_triggers: np.ndarray,
    arr_parcels: np.ndarray = None,
    nodata=None,
    mode=1,
):
    """Expand grass strips based on neighbouring trigger pixels.

    This algorithm expands grass strips with one neighbouring pixel when a neighbouring
    is a trigger pixel. The expansion can be bound within one unique parcel (optional).

    Parameters
    ----------
    arr_grass_strips: numpy.ndarray
        Array with grass strips. Every contiguous grass strips needs to be identified
        with an unique id as array value.
    arr_triggers: numpy.ndarray
        Binary array with 1's when a pixel is a trigger. Nodata values are allowed.
    arr_parcels: numpy.ndarray, default None
        See :func:`pywatemsedem.grasstrips.core_expand_grass_strips_with_triggers`
    nodata: float, default None
        See :func:`pywatemsedem.grasstrips.core_expand_grass_strips_with_triggers`
    mode: int, default 1
        See :func:`pywatemsedem.grasstrips.core_expand_grass_strips_with_triggers`.
        Only for computing neighbours grass strips.

    Returns
    -------
    arr_grass_strips: np.ndarray
        See :func:`pywatemsedem.grassstrips.core_expand_grass_strips_with_triggers`

    Examples
    --------
    >>> # triggers (for instance river, road)
    >>> arr_triggers = np.array([[1, 0, 0, 0],
    >>>                        [0, 1, 0, 0],
    >>>                        [0, 0, 1, 0],
    >>>                        [0, 0, 0, 1]])
    >>> # parcel ids
    >>> arr_parcels_id = np.array([[0, 3, 3, 3],
    >>>                            [1, 0, 2, 2],
    >>>                            [1, 1, 0, 2],
    >>>                            [1, 1, 1, 0]])
    >>> # input gras id's
    >>> arr_grass_strips = np.array([[0, 0, 0, 0],
    >>>                            [0, 0, 0, 0],
    >>>                            [0, 0, 0, 1],
    >>>                            [0, 0, 0, 1]])
    >>> # execute
    >>> arr_out = expand_grass_strips_with_triggers(
    >>>    arr_grass_strips, arr_triggers, arr_parcels_id)

    Notes
    -----
    1. For algorithm description, see
    :func:`pywatemsedem.grasstrips.core_expand_grass_strips_with_triggers`

    For visual example, see image below:

    .. image:: /_static/png/expand_grass_strips_at_infr_parcel_boundary.png
      :width: 600


    2. All input arrays should have the same nodata value in their raster.
    """
    if arr_triggers.shape != arr_grass_strips.shape:
        msg = (
            f"Input grass strips array {arr_grass_strips.shape} has a different "
            f"size than the input triggers array {arr_triggers.shape}."
        )
        raise IOError(msg)

    if set(arr_triggers.ravel()) != {0, 1}:
        msg = "Input trigger can only contain 0's and 1's (and no data) values."
        raise ValueError(msg)

    # remove grass strips from triggers
    arr_grass_strips_nt = arr_grass_strips.copy()
    arr_grass_strips_nt[arr_triggers == 1] = 0

    # add boundaries to cope with boundary condition
    arr_grass_strips = add_boundary_rows_cols_to_arr(arr_grass_strips)
    arr_grass_strips_nt = add_boundary_rows_cols_to_arr(arr_grass_strips_nt)
    arr_triggers = add_boundary_rows_cols_to_arr(arr_triggers)
    if arr_parcels is not None:
        arr_parcels = add_boundary_rows_cols_to_arr(arr_parcels)

    # compute for every pixel how many pixels are trigger pixels
    arr_triggers_neighbours = compute_number_of_non_zero_neighbours(
        arr_triggers, nodata=nodata
    )
    # compute for every pixel how many pixels are gras strip pixels
    arr_grass_strips_neighbours = compute_number_of_non_zero_neighbours(
        arr_grass_strips_nt, nodata=nodata, mode=mode
    )

    # run expansion algorithm
    arr_grass_strips = core_expand_grass_strips_with_triggers(
        arr_grass_strips,
        arr_grass_strips_neighbours,
        arr_triggers_neighbours,
        arr_parcels,
        nodata=nodata,
    )

    return arr_grass_strips[1:-1, 1:-1]


def compute_number_of_non_zero_neighbours(arr, nodata=None, mode=1):
    """Compute number of non-zero neighbour (eight) elemnts for every
    array element.

    This algorithm computes how many neighbours of a pixel are not equal to zero (and
    not equal to nodata.)

    Parameters
    ----------
    arr: numpy.ndarray
        Input array with valid values: (nodata, 0, >0)
        Nodata values are considered as 0-values. >0 values are considerd as 1-values
    nodata: float, default None
        Values to ignore, if equal to None this is not considered.
    mode: int
        Mode of expansion:

        1: Consider ordinal + cardinal direction.
        2: Cardinal direction.
        3: Ordinal direction.

    Returns
    -------
    arr_out: numpy.ndarray
        Array with for each element the number of non-zero neighbours.

    Notes
    -----
    1. Nodata values are not considered!
    2. Boundaries are set to zero.
    3. Grass strips within trigger pixels are not considered.
    """
    arr_out = arr.copy()
    if nodata is not None:
        arr_out[arr_out == nodata] = 0
    arr_out = arr_out.astype(bool)

    if mode == 1:
        weights = np.ones((3, 3))
    elif mode == 2:
        weights = np.array([[1, 0, 1], [0, 1, 0], [1, 0, 1]])
    elif mode == 3:
        weights = np.array([[0, 1, 0], [1, 1, 1], [0, 1, 0]])

    else:
        msg = (
            f"Mode for searching neighbours {mode} not known, please select 1 "
            f"(cardinal + ordinal), 2 (ordinal direction) or 3 (cardinal direction)."
        )
        raise KeyError(msg)

    arr_out[1:-1, 1:-1] = signal.convolve2d(arr_out, weights, "valid")

    # boundaries
    arr_out[0, :] = 0
    arr_out[-1, :] = 0
    arr_out[:, 0] = 0
    arr_out[:, -1] = 0

    arr_out[(arr != 0) & (arr != nodata)] = 0

    return arr_out


def add_boundary_rows_cols_to_arr(arr):
    """Add boundary rows and columns to an array, set values to zero.

    Parameters
    ----------
    arr: numpy.ndarray

    Returns
    -------
    numpy.ndarray
        With two rows added to the top and bottom, and two columns added to the left
        and right
    """
    shape = arr.shape
    _arr = np.zeros((shape[0] + 2, shape[1] + 2))
    _arr[1:-1, 1:-1] = arr

    return _arr


def core_expand_grass_strips_with_triggers(
    arr_grass_strips,
    arr_grass_strips_neighbours,
    arr_triggers_neighbours,
    arr_parcels=None,
    nodata=None,
):
    """Grass strip expansion algorithm based on raster inputs.

    The grass strips expansion algorithm executes following steps:

    1. Expand the grass strips with one pixel. This implies that adjacent (ordinal
       and/or cardinal directions, depending on mode, see
       :func:`pywatemsedem.grasstrips.compute_number_of_non_zero_neighbours`) pixels to
       a grass strip pixel are classified as grass strip pixels.

    2. The added grass strip pixels can only be retained if following rules are
       satisfied:

        - The added grass strip pixels is in the direct vicinity of (one of 8
          neighbours) a trigger pixel (river, infrastructure pixel).
        - (optional) The added grass strip pixels can only be withheld when it belongs
          to a parcel to which the original grass strip was part of.

    Parameters
    ----------
    arr_grass_strips: numpy.ndarray
        Array with unique id's per grass strips
    arr_grass_strips_neighbours:
        See :func:`pywatemsedem.grasstrips.compute_number_of_non_zero_neighbours` applied
        on grass strips array.
    arr_triggers_neighbours: nump.ndarray
        See :func:`pywatemsedem.grasstrips.compute_number_of_non_zero_neighbours` applied
        on triggers array.
    arr_parcels: nump.ndarray, default None
        Parcel ids raster. Pixel belonging to one parcel share the same unique id. The
        value zero indicates that no parcel is present. If None, the expansion is not
        limited to the boundaries of a parcel.
    nodata: float, default None
        Nodata value of all rasters

    Returns
    -------
    arr_grass_strips: numpy.ndarray
        Grass strips array with grass strips expanded at the
        trigger boundary edge cases.

    Notes
    -----
    1. This algorithm is not able to make a distinction between two separate grass
       strip in one parcel.

    2. For the expansion of the grass strips, the source grass pixels overlapping with
       the triggers are not considered. This implies that if a specific gras strip
       completely overlaps with triggers, no expansion will be done.
    """
    for id_grass_strip in np.unique(arr_grass_strips):

        if (id_grass_strip != 0) & (id_grass_strip != nodata):

            if arr_parcels is None:
                cond = (arr_grass_strips_neighbours > 0) & (arr_triggers_neighbours > 0)
                arr_grass_strips[cond] = id_grass_strip
            else:
                un_parcels_id = np.unique(
                    arr_parcels[(arr_grass_strips == id_grass_strip)]
                )
                un_parcels_id = un_parcels_id[un_parcels_id > 0]

                for parcel_id in un_parcels_id:
                    cond = (
                        (arr_parcels == parcel_id)
                        & (arr_grass_strips_neighbours > 0)
                        & (arr_triggers_neighbours > 0)
                    )
                    arr_grass_strips[cond] = id_grass_strip

    return arr_grass_strips


def create_grassstrips_cnws(
    arr_grass_strips_ids,
    arr_river,
    arr_infr,
    profile,
    arr_parcels=None,
    expand_grass_strips=False,
):
    """Create grass strips raster for WaTEM/SEDEM

    This function create the grass strips raster that is used as input creating the
    composite WaTEM/SEDEM parcels raster (see :ref:`here <watemsedem:prcmap>`). The
    output of this function provides a raster with -6-values for grass strip pixels.

    Parameters
    ----------
    arr_grass_id: numpy.ndarray
        Grass strips id's-raster. Pixel belonging to one grass strip share the same
        unique id. Other values should be have nodata-value described in the parameter
        profile.
    arr_river: numpy.ndarray
        River raster. River pixels should have the value differing from 0 or nodata
        (defined in profile).
    arr_infr: numpy.ndarray
        Infrastructure raster. Infrastructure pixels should have the value differing
        from 0 or nodata (defined in profile).
    arr_parcels: numpy.ndarray, default None
        Parcel ids raster. Pixel belonging to one parcel share the same unique id. The
        value zero indicates that no parcel is present. If None, the expansion is not
        limited to the boundaries of a parcel.
    profile: dict
        Rasterio profile, see
        :attr:`pywatemsedem.geo.rasterproperties.RasterProperties.rasterio_profile`. All the
        input rasters must have the same nodata-value, defined in profile.
    expand_grass_strips: bool, default False
        Use expand grass strips algorithm with rivers as triggers, see
         :func:`pywatemsedem.grasstrips.expand_grass_strips_with_triggers`

    Returns
    -------
    arr_grass_strips_ids: numpy.ndarray
        Expanded grass strips id's-raster.
    arr_grass: numpy.ndarray
        -6-value-raster of grass strips id's-raster, formatted according to composite
         WaTEM/SEDEM parcels raster (see :ref:`here <watemsedem:prcmap>`).

    Notes
    -----
    All the input rasters must have the same nodata-value, defined in profile.
    """
    if expand_grass_strips:
        arr_river[(arr_river == profile["nodata"])] = 0
        arr_infr[(arr_infr == profile["nodata"])] = 0
        arr_triggers = arr_infr + arr_river
        arr_triggers[(arr_triggers != 0)] = 1
        # arr_gras[arr_gras==profile["nodata"]] = 0

        arr_grass_strips_ids = expand_grass_strips_with_triggers(
            arr_grass_strips_ids,
            arr_triggers,
            arr_parcels=arr_parcels,
            nodata=profile["nodata"],
            mode=1,
        )
    arr_grass = arr_grass_strips_ids.copy()
    arr_grass[arr_grass != profile["nodata"]] = -6

    return arr_grass_strips_ids, arr_grass


def get_neighbour_grass_strips_ids_array(
    vct_grass_strips, rst_params, width_polygon=20
):
    gdf = gpd.read_file(vct_grass_strips)
    gdf["buffer"] = gdf.buffer(width_polygon)
    gdf = gdf[["NR", "buffer"]]
    gdf.rename(columns={"buffer": "geometry"}, inplace=True)
    gdf = gpd.GeoDataFrame(gdf, geometry="geometry")
    gdf.to_file(vct_grass_strips.parent / (vct_grass_strips.stem + "_buffer" + ".shp"))
    rst_out = "neighbour_grass_strips_ids_array.tiff"
    vct_to_rst_field(
        vct_grass_strips, Path(rst_out), rst_params, "NR", alltouched=True, dtype=None
    )


def extract_grass_strips_from_parcels(vct_parcels, year, resmap=Path.cwd(), tag=""):
    """Extract grass buffers from parcels using the Thinnes criterium

    Extract grass buffers from parcels using the Thinnes
    criterium (>0.3) of a given year.

    ! EXPERIMENTAL - NOT IN USE YET !

    Parameters
    ----------
    year : int
        Year for which to extract the grass buffers.
    """
    grass_strips = resmap / f"graspercelen_{year}_{tag}.shp"
    d = {}

    try:
        gdf = gpd.read_file(vct_parcels)
    except Exception:
        msg = f"could not open {vct_parcels}"
        raise IOError(msg)
    else:
        gdf = gdf.loc[gdf["Grasbuffer"] == 1]
        gdf["opp"] = gdf.area
        gdf["perimeter"] = gdf.length

        gdf = gdf.loc[gdf.length**2 >= 16 * gdf.area]
        gdf["width"] = (gdf.length - np.sqrt((gdf.length**2) - (16 * gdf.area))) / 4

        gdf["Aspect"] = (gdf["width"] ** 2) / gdf.area
        gdf["Thinnes"] = 4 * np.pi * gdf.area / (gdf.length**2)
        gdf = gdf.loc[gdf["Thinnes"] < 0.5]
        gdf = gdf.loc[gdf["width"] < 25]
        gdf = gdf.loc[gdf["Aspect"] < 0.25]
        if not gdf.empty:
            d["gras"] = grass_strips
            gdf.to_file(grass_strips)
        vct_grass_strips_parcels = d

    return vct_grass_strips_parcels


def create_grass_strips_from_line_string(
    line_string, polygons=None, width=20, width_polygon=20
):
    """Add bank grass strips

    Adds bank grass strips with a standard length of 20m, along input LineStrings. If
    required, clip by polygons.

    Parameters
    ----------
    line_string: geopandas.GeoDataFrame
        Geopandas dataframe with LineString
    polygons: geopandas.GeoDataFrame, default None
        Geopandas dataframe with Polygon
    width: int, optional, default 20
        Width (m) of the bank grass strip stored in the attribute table.
    width_polygon: int, optional, default 20
        width (m) of the bank grass strip polygon

    Returns
    -------
    grass_strips: geopandas.GeoDataFrame

        - *width* (float): width of grass strips
        - *scale_ktc* (float): scale ktc, see
          :func:`pywatemsedem.grassttips.scale_ktc_with_grass_strip_width`


    Notes
    -----
    The `width` is the estimated true width of the grass strip, whereas the
    `width polygon` the used width is for mapping the shape to a raster. Typically,
    this `width polygon` is set to the resolution to ensure all LineString-surrounding
    pixels are mapped as grass strips. Note that the `width polygon` is not the
    estimated true width, it is a mapping (shape => raster) width.
    """

    line_string["buffer"] = line_string.buffer(width_polygon)
    grass_strips = line_string.drop(columns="geometry")
    grass_strips["width"] = width
    grass_strips["source"] = "bankgrassstrip"
    grass_strips["scale_ktc"] = 1
    grass_strips.rename(columns={"buffer": "geometry"}, inplace=True)
    grass_strips = gpd.GeoDataFrame(grass_strips, geometry="geometry")

    # intersect over years
    # if no intersection with parcels: assign same bank grass strip for every
    # year
    if polygons is not None:
        tmp_bankgrasstrips = tempfile.NamedTemporaryFile(
            prefix=PREFIX_TEMP, suffix=".shp"
        ).name
        grass_strips.to_file(tmp_bankgrasstrips)
        tmp_polygons = tempfile.NamedTemporaryFile(
            prefix=PREFIX_TEMP, suffix=".shp"
        ).name
        polygons.to_file(tmp_polygons)
        tmp_bankstrips_polygons = tempfile.NamedTemporaryFile(
            prefix=PREFIX_TEMP, suffix=".shp"
        ).name
        saga_intersection(
            tmp_bankgrasstrips,
            tmp_polygons,
            tmp_bankstrips_polygons,
        )
        grass_strips = gpd.read_file(tmp_bankstrips_polygons)
        # add dissolve of grass strips here
        grass_strips = grass_strips.dissolve("NR")
        grass_strips["width"] = width
        grass_strips["scale_ktc"] = 1
        clean_up_tempfiles(tmp_bankgrasstrips, "shp")
        clean_up_tempfiles(tmp_polygons, "shp")
        clean_up_tempfiles(tmp_bankstrips_polygons, "shp")

    return grass_strips.reset_index()[["width", "scale_ktc", "geometry"]]



import geopandas as gpd
import pytest
from conftest import geodata
from shapely.geometry import Polygon

from pywatemsedem.defaults import SAGA_FLAGS
from pywatemsedem.geo.utils import any_equal_element_in_vector, execute_subprocess


@pytest.mark.skip(reason="Unvalidated")
class TestExecuteSubprocess:
    """Test execute subprocess"""

    def test_execute_subprocess(self, tmp_path):
        """Test execute subprocess

        Parameters
        ----------
        tmp_path
        """

        # input
        vct_out = tmp_path / "test.shp"

        # test if a 'random' command line runs
        cmd_args = [
            "saga_cmd",
            SAGA_FLAGS,
            "shapes_grid",
            "6",
            "-GRID",
            str(geodata.rst_mask),
        ]
        cmd_args += ["-POLYGONS", str(vct_out), "-CLASS_ALL", "1", "-SPLIT", "0"]
        assert execute_subprocess(cmd_args)

    def test_execute_subprocess_error(self, tmp_path):
        """Test execute subprocess errors

        Parameters
        ----------
        tmp_path
        """
        # input
        vct_out = tmp_path / "test.shp"

        # catch error: in this case, command file because module not found
        cmd_args = [
            "saga_cmd",
            SAGA_FLAGS,
            "shapes_gridzzzz",
            "6",
            "-GRID",
            str(geodata.rst_mask),
        ]
        cmd_args += ["-POLYGONS", str(vct_out), "-CLASS_ALL", "1", "-SPLIT", "0"]
        with pytest.raises(OSError) as excinfo:
            execute_subprocess(cmd_args)
        assert "Error: select a library" in str(excinfo.value)


class TestAnyEqualVector:
    """Test if any element in vector is equal"""

    s1 = gpd.GeoSeries(
        [Polygon([(0, 0), (2, 2), (0, 2)]), Polygon([(0, 0), (1, 2), (0, 2)])]
    )

    s2 = gpd.GeoSeries(
        [Polygon([(0, 0), (2, 2), (0, 2)]), Polygon([(0, 0), (1, 2), (0, 3)])]
    )

    s3 = gpd.GeoSeries(
        [Polygon([(0, 0), (2, 2), (0, 3)]), Polygon([(0, 0), (1, 2), (0, 3)])]
    )

    def test_any_equal_element_in_vector(self):
        """Check if any element in the left series is equal to the right element"""
        assert any_equal_element_in_vector(self.s1, self.s2) is True
        assert any_equal_element_in_vector(self.s1, self.s3) is False



from pathlib import Path

import numpy as np
import pandas as pd
import pytest
from pytest import approx

from pywatemsedem.io.modeloutput import (
    Modeloutput,
    check_segment_edges,
    compute_efficiency_buffers,
    identify_rank_sediment_loads,
)


@pytest.mark.skip(
    reason="test gives saga-issues in conda pipeline, ommited, raised as"
    " issue in saga"
)
def test_modelouput():
    """Testing the modeloutput class"""
    folder = Path(r"tests/io/data")
    filepath_out = folder / "model_out"
    filepath_in = folder / "model_in"
    P_ex = filepath_in / "pfactor.rst"
    example = Modeloutput(P_ex, resolution=20, epsg=31370, nodata=-9999)
    # sediout
    example.sediout = filepath_out / "SediOut_kg.rst"
    example.sediout.plot()
    # example.sediout.hv_plot()
    # routing
    example.routing = filepath_out / "routing.txt"
    landuse_path = filepath_in / "perceelskaart.rst"
    sediout_path = filepath_out / "SediOut_kg.rst"
    example.make_routing_vector(landuse_path, sediout_path)
    # example.routing.plot()
    # routing_missing
    example.routing_missing = filepath_out / "routing_missing.txt"
    example.make_routing_vector(landuse_path, sediout_path, routing_missing=True)
    example.routing_missing.plot()
    # ls
    example.ls = filepath_out / "LS.rst"
    example.ls.plot()
    # example.ls.hv_plot()
    # slope
    example.slope = filepath_out / "SLOPE.rst"
    example.slope.plot()
    # example.slope.hv_plot()
    # uparea
    example.uparea = filepath_out / "UPAREA.rst"
    example.uparea.plot()
    # example.uparea.hv_plot()
    # total sediment
    example.total_sediment = filepath_out / "Total sediment.txt"
    # sewer in
    example.sewer_in = filepath_out / "sewer_in.rst"
    example.sewer_in.plot()
    # example.sewer_in.hv_plot()
    # SediExport
    example.sedi_export = filepath_out / "SediExport_kg.rst"
    example.sedi_export.plot()
    # example.sedi_export.hv_plot()
    # SediIn
    example.sedi_in = filepath_out / "SediIn_kg.rst"
    example.sedi_in.plot()
    # example.sedi_in.hv_plot()
    # Capacity
    example.capacity = filepath_out / "Capacity.rst"
    example.capacity.plot()
    # example.capacity.hv_plot()
    # RUSLE
    example.rusle = filepath_out / "RUSLE.rst"
    example.rusle.plot()
    # example.rusle.hv_plot()


def test_compute_efficiency_buffers():
    """Compute efficiency buffers"""

    exp_sediin = np.array([5928.877930, 4209.729492])
    exp_sediout = np.array([1482.219482, 1052.4324])
    buff_sed = np.array([4446.658203, 3157.297])

    folder = Path(r"tests/io/data")
    filepath_out = folder / "model_out"
    filepath_in = folder / "model_in"

    df = compute_efficiency_buffers(
        filepath_in / "buffers.rst",
        filepath_out / "SediIn_kg.rst",
        filepath_out / "SediOut_kg.rst",
    )

    np.testing.assert_allclose(df["sediout"], exp_sediout, atol=1e-2)
    np.testing.assert_allclose(df["sediin"], exp_sediin, atol=1e-2)
    np.testing.assert_allclose(df["buff_sed"], buff_sed, atol=1e-2)


@pytest.mark.parametrize(
    "threshold,n_ranks,sum_sediment_load,mean_sediment_load",
    [
        (50, 12, 337717.72, 28143.142578125),
        (20, 2, 100766.0, 50383.0),
    ],
)
def test_identify_rank_sediment_loads(
    threshold, n_ranks, sum_sediment_load, mean_sediment_load, tmp_path
):
    """Test function for :func:`pywatemsedem.pywatemsedem.process_output.map_rank_sediment_loads``

    Parameters
    ----------
    threshold: float
        See :func:`pywatemsedem.pywatemsedem.process_output.analyse_cumulative_sediexport`
    n_ranks: float
        Number of records (ranks) for threshold.
    sum_sediment_load: float
        Total sediment load for number of records.
    mean_sediment_load: float
        Mean sediment load for number of records
    tmp_path
    """
    tag = "rank"
    folder = Path(r"tests/io/data")

    filepath_out = folder / "model_out" / "SediExport_kg.rst"

    df_export, _ = identify_rank_sediment_loads(
        filepath_out,
        threshold,
        tmp_path / (tag + ".rst"),
        rst_endpoints=folder / "model_out" / "sewer_in.rst",
    )
    df_export = df_export[df_export["class"] != -9999]

    assert df_export.shape[0] == n_ranks
    assert np.sum(df_export["sediexport"]) == approx(sum_sediment_load, abs=1)
    assert np.mean(df_export["sediexport"]) == approx(mean_sediment_load, abs=1)


def test_check_segment_edges():
    """Test function for check test_check_segment_edges."""

    # normal, valid input: from the documentation
    arr = np.array(
        [
            [0, 0, 0, 0, 0, 5, 0, 0, 0, 0],
            [0, 0, 0, 0, 0, 5, 0, 0, 0, 0],
            [0, 0, 0, 0, 0, 5, 0, 0, 7, 0],
            [0, 0, 0, 0, 4, 3, 7, 7, 0, 0],
            [0, 0, 0, 4, 0, 3, 0, 0, 0, 0],
            [0, 0, 4, 0, 0, 3, 0, 0, 0, 0],
            [0, 4, 0, 0, 0, 3, 0, 0, 0, 0],
            [0, 0, 0, 0, 1, 0, 2, 0, 0, 0],
            [0, 0, 0, 1, 0, 0, 0, 2, 0, 0],
            [0, 0, 0, 1, 0, 0, 2, 0, 0, 0],
            [0, 0, 1, 0, 0, 6, 2, 0, 0, 0],
        ]
    )

    up_edges = pd.DataFrame(
        {
            "line_id": [3, 3, 5, 5, 5, 5, 5, 5, 6, 5],
            "upstream_line": [1, 2, 4, 3, 2, 1, 6, 7, 2, 1],
            "proportion": [1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0],
        }
    )

    adj_edges = pd.DataFrame({"from": [1, 2, 3, 4, 6, 7], "to": [3, 3, 5, 5, 2, 5]})

    # test 1: not expected any warnings or filtering!
    adj_edges_adj, up_edges_adj, filt = check_segment_edges(adj_edges, up_edges, arr)
    np.testing.assert_equal(filt, False)
    np.testing.assert_allclose(
        adj_edges_adj["from"],
        adj_edges["from"],
    )
    np.testing.assert_allclose(
        adj_edges_adj["to"],
        adj_edges["to"],
    )
    np.testing.assert_allclose(up_edges_adj["line_id"], up_edges["line_id"])
    np.testing.assert_allclose(up_edges_adj["upstream_line"], up_edges["upstream_line"])

    # test 2: add non valid edges in adj_edges in from series
    add = pd.DataFrame.from_dict({"from": [9], "to": [1]})
    adj_edges_frombis = pd.concat([adj_edges, add], ignore_index=True)
    adj_edges_adj, up_edges_adj, filt = check_segment_edges(
        adj_edges_frombis, up_edges, arr
    )
    np.testing.assert_equal(filt, True)
    # expeced to filter out what is not desired!
    np.testing.assert_allclose(
        adj_edges_adj["from"],
        adj_edges["from"],
    )
    np.testing.assert_allclose(
        adj_edges_adj["to"],
        adj_edges["to"],
    )
    np.testing.assert_allclose(up_edges_adj["line_id"], up_edges["line_id"])
    np.testing.assert_allclose(up_edges_adj["upstream_line"], up_edges["upstream_line"])

    # test 3: add non valid edges in adj_edges in to series
    add = pd.DataFrame.from_dict({"from": [1], "to": [9]})
    adj_edges_tobis = pd.concat([adj_edges, add], ignore_index=True)
    adj_edges_adj, up_edges_adj, filt = check_segment_edges(
        adj_edges_tobis, up_edges, arr
    )
    np.testing.assert_equal(filt, True)
    # expeced to filter out what is not desired!
    np.testing.assert_allclose(
        adj_edges_adj["from"],
        adj_edges["from"],
    )
    np.testing.assert_allclose(
        adj_edges_adj["to"],
        adj_edges["to"],
    )
    np.testing.assert_allclose(up_edges_adj["line_id"], up_edges["line_id"])
    np.testing.assert_allclose(up_edges_adj["upstream_line"], up_edges["upstream_line"])

    # est 4: add no valid edge in up_edges in edge series
    add = pd.DataFrame.from_dict({"line_id": [9], "upstream_line": [1]})
    up_edges_edgebis = pd.concat([up_edges, add], ignore_index=True)

    adj_edges_adj, up_edges_adj, filt = check_segment_edges(
        adj_edges, up_edges_edgebis, arr
    )
    np.testing.assert_equal(filt, True)
    # expeced to filter out what is not desired!
    np.testing.assert_allclose(
        adj_edges_adj["from"],
        adj_edges["from"],
    )
    np.testing.assert_allclose(
        adj_edges_adj["to"],
        adj_edges["to"],
    )
    np.testing.assert_allclose(up_edges_adj["line_id"], up_edges["line_id"])
    np.testing.assert_allclose(up_edges_adj["upstream_line"], up_edges["upstream_line"])

    # test 5: add no valid edge in up_edges in upstream_edge series
    add = pd.DataFrame.from_dict({"line_id": [1], "upstream_line": [9]})
    up_edges_upedgebis = pd.concat([up_edges, add], ignore_index=True)

    adj_edges_adj, up_edges_adj, filt = check_segment_edges(
        adj_edges, up_edges_upedgebis, arr
    )
    np.testing.assert_equal(filt, True)
    # expeced to filter out what is not desired!
    np.testing.assert_allclose(
        adj_edges_adj["from"],
        adj_edges["from"],
    )
    np.testing.assert_allclose(
        adj_edges_adj["to"],
        adj_edges["to"],
    )
    np.testing.assert_allclose(up_edges_adj["line_id"], up_edges["line_id"])
    np.testing.assert_allclose(up_edges_adj["upstream_line"], up_edges["upstream_line"])

    # test 6: a nan input
    up_edges_nan = pd.DataFrame(
        {
            "line_id": [3, 3, 5, 5, 5, 5, 5, 5, 6, np.nan],
            "upstream_line": [1, 2, 4, 3, 2, 1, 6, 7, 2, 1],
            "proportion": [1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0],
        }
    )
    # expected to delete the last row
    up_edges_nan_exp = pd.DataFrame(
        {
            "line_id": [3, 3, 5, 5, 5, 5, 5, 5, 6],
            "upstream_line": [1, 2, 4, 3, 2, 1, 6, 7, 2],
            "proportion": [1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0],
        }
    )
    adj_edges_adj, up_edges_adj, filt = check_segment_edges(
        adj_edges, up_edges_nan, arr
    )
    np.testing.assert_equal(filt, True)
    # expeced to filter out what is not desired!
    np.testing.assert_allclose(
        adj_edges_adj["from"],
        adj_edges["from"],
    )
    np.testing.assert_allclose(
        adj_edges_adj["to"],
        adj_edges["to"],
    )
    np.testing.assert_allclose(up_edges_adj["line_id"], up_edges_nan_exp["line_id"])
    np.testing.assert_allclose(
        up_edges_adj["upstream_line"], up_edges_nan_exp["upstream_line"]
    )

    # test 7: a string input
    up_edges_string = pd.DataFrame(
        {
            "line_id": [3, 3, 5, 5, 5, 5, 5, 5, 6, "what, a string?"],
            "upstream_line": [1, 2, 4, 3, 2, 1, 6, 7, 2, 1],
            "proportion": [1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0],
        }
    )
    adj_edges_adj, up_edges_adj, filt = check_segment_edges(
        adj_edges, up_edges_string, arr
    )
    up_edges_adj = up_edges_adj.astype("int64")  # change to this to compare values!!
    # otherwise dtype = object when string is used as input!
    np.testing.assert_equal(filt, True)
    # expeced to filter out what is not desired!
    np.testing.assert_allclose(
        adj_edges_adj["from"],
        adj_edges["from"],
    )
    np.testing.assert_allclose(
        adj_edges_adj["to"],
        adj_edges["to"],
    )
    np.testing.assert_allclose(up_edges_adj["line_id"], up_edges_nan_exp["line_id"])
    np.testing.assert_allclose(
        up_edges_adj["upstream_line"], up_edges_nan_exp["upstream_line"]
    )



"""Test functions for postprocessing functions"""
import numpy as np
import pandas as pd
import pytest
from conftest import postprocess
from numpy.testing import assert_almost_equal

from pywatemsedem.postprocess import (
    compute_efficiency_grass_strips,
    compute_netto_erosion_parcels,
    read_filestructure,
)


@pytest.mark.parametrize("sep", [",", ";"])
def test_read_filestructure(sep):
    """Test function for reading filestructure file from pywatemsedem package

    The function is tested by using a correct delimiter (sep=",") which leads to a
    succesfull load of the text file (","-delimited), and by using a wrong delimiter
    (sep=";") which leads to a fail of loading the text file.

    Parameters
    ----------
    sep: str
        Delimiter.
    """
    if sep == ";":
        with pytest.raises(KeyError) as excinfo:
            read_filestructure(sep=sep)
        assert "DataFrame should contain " in str(excinfo.value)
    else:
        df = read_filestructure(sep=sep)
        assert len(df) > 0


def test_get_grass_strips_statistics():
    """Test function for get_gras_strips_statistics for individual grass strips.
    This test function compares the arrays gras id, and coupled sedi in and out."""

    # run function
    _, _, df = compute_efficiency_grass_strips(
        postprocess.txt_routing,
        postprocess.rst_grasstrips_id,
        postprocess.rst_prclskrt,
        postprocess.rst_sediout,
    )

    df_test = pd.read_csv(postprocess.txt_gras_efficiency)

    assert_almost_equal(df["gras_id_target"].values, df_test["gras_id_target"].values)
    assert_almost_equal(df["gras_id_source"].values, df_test["gras_id_source"].values)
    assert_almost_equal(df["npixels_t"].values, df_test["npixels_t"].values)
    assert_almost_equal(df["eSTE"].values, df_test["STE"].values)
    assert_almost_equal(df["sediin"].values, df_test["sediin"].values)


def test_compute_netto_erosion_parcels():
    """Test function for computing netto erosion for individual parcels. It
    used two parcels rasters, one with a int16 codation (limited number of
    parcels) and one with a float64. The assert will test whether the sum, average
    and standard deviation on the netto erosion per parcel is equal for all parcels
    in the input data."""

    # script
    df_output, _ = compute_netto_erosion_parcels(
        postprocess.rst_prclskrt,
        postprocess.rst_watereros,
        postprocess.rst_rasterized_prc_shp,
        flag_write=True,
    )
    file_name = "average_netto_erosion"
    np.testing.assert_allclose(
        df_output[file_name],
        np.loadtxt(postprocess.txt_average_netto_erosion),
        atol=1e-2,
        rtol=1e-02,
    )
    file_name = "std_dev_netto_erosion"
    np.testing.assert_allclose(
        df_output[file_name],
        np.loadtxt(postprocess.txt_std_dev_netto_erosion),
        atol=1e-2,
        rtol=1e-02,
    )
    file_name = "sum_netto_erosion"
    np.testing.assert_allclose(
        df_output[file_name],
        np.loadtxt(postprocess.txt_sum_netto_erosion),
        atol=1e-2,
        rtol=1e-02,
    )
    file_name = "area_parcel"
    np.testing.assert_allclose(
        df_output[file_name],
        np.loadtxt(postprocess.txt_area_parcel),
        atol=1e-2,
        rtol=1e-02,
    )



import numpy as np
import pytest

from pywatemsedem.cfactor import reduce_cfactor_with_source_oriented_measures


class TestReduceCfactorWithSourceOrientedMeasures:
    """Class to test behaviour reduce C-factor reduction with source oriented
    measures"""

    c_factor = np.array([0.2, 0.8])
    c_reduction = np.array([0.5, 0.7])

    def test_reduce_values(self, recwarn):
        """Test reducing of value of C_factor with C_reduction. In addition test
        warning message from reduction.
        """
        c_factor_out = reduce_cfactor_with_source_oriented_measures(
            self.c_factor, self.c_reduction, True
        )
        np.testing.assert_almost_equal(c_factor_out, np.array([0.2 * 0.5, 0.8 * 0.3]))
        w = recwarn.pop(UserWarning)
        assert (
            str(w.message)
            == "Implementing source-oriented measures: reducing C-factor with "
            "value defined in '1 - C_reduction'."
        )

    def test_no_use_of_source_oriented_measures(self):
        """Test if the same C-factor valeus are obtained when not using
        reduction-functionality."""
        c_factor_out = reduce_cfactor_with_source_oriented_measures(
            self.c_factor, self.c_reduction, False
        )
        np.testing.assert_almost_equal(c_factor_out, np.array([0.2, 0.8]))

    def test_out_of_range_values(self):
        """Test whether an error is returned when the reduction-flag is on, and no
        C_reduction column is defined."""
        c_factor = np.array([0.2, 1.8])
        err_msg = "Values of c-factor should be between 0 and 1."
        with pytest.raises(IOError, match=err_msg):
            reduce_cfactor_with_source_oriented_measures(
                c_factor, self.c_reduction, True
            )

        c_reduction = np.array([0.2, 1.8])
        err_msg = "Values of c-reduction should be between 0 and 1."
        with pytest.raises(IOError, match=err_msg):
            reduce_cfactor_with_source_oriented_measures(
                self.c_factor, c_reduction, True
            )

        c_factor = np.array([0.2, np.nan])
        err_msg = "Values of c-factor should be between 0 and 1."
        with pytest.raises(IOError, match=err_msg):
            reduce_cfactor_with_source_oriented_measures(
                c_factor, self.c_reduction, True
            )

        c_reduction = np.array([0.2, np.nan])
        err_msg = "Values of c-reduction should be between 0 and 1."
        with pytest.raises(IOError, match=err_msg):
            reduce_cfactor_with_source_oriented_measures(
                self.c_factor, c_reduction, True
            )



import logging
from dataclasses import dataclass

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
from matplotlib import cm, colors
from sklearn import linear_model
from sklearn.metrics import r2_score

logger = logging.getLogger(__name__)


@dataclass
class Calibration:
    """Data class for calibration parameters.

    Data class used to store all commonly used variables for a calibration run.

    Parameters
    ----------
    ktc_low_min: float
        Lower boundary for ktc low value
    ktc_low_max: float
        Upper boundary for ktc low value
    ktc_high_min: float
        Lower boundary for ktc high value
    ktch_high_max: float
        Upper boundary for ktc high value
    steps: int
        Amount of steps between lower and upper ktc low/high values

    """

    ktc_low_min: float
    ktc_low_max: float
    ktc_high_min: float
    ktc_high_max: float
    steps: int

    def __post_init__(self):

        self.check_input()
        self.define_parameter_grid_resolution()
        self.define_parameter_grid()
        self.df_template = _make_template_df(self.arr_ktc_low, self.arr_ktc_high)

    def check_input(self):
        """Run a number of checks on input data of data class"""
        if self.ktc_low_min > self.ktc_low_max:
            msg = "ktc low min > ktc low max"
            raise ValueError(msg)

        if self.ktc_high_min > self.ktc_high_max:
            msg = "ktc high min > ktc high max"
            raise ValueError(msg)

        if self.steps <= 0:
            msg = "the amount of steps must be larger than zero"
            raise ValueError(msg)

        if self.ktc_low_min < 0:
            msg = "ktc low min must be higher than 0"
            raise ValueError(msg)

        if self.ktc_high_min < 0:
            msg = "ktc high min must be higher than 0"
            raise ValueError(msg)

    def define_parameter_grid_resolution(self):
        """Define the resolution of the parameter grid (2D, x and y resolution)."""
        self.stepresolution_high = (self.ktc_high_max - self.ktc_high_min) / self.steps
        self.stepresolution_low = (self.ktc_low_max - self.ktc_low_min) / self.steps

    def define_parameter_grid(self):
        """Defne parameter grid for ktc low and high"""
        self.arr_ktc_low = np.arange(
            self.ktc_low_min,
            self.ktc_low_max + self.stepresolution_low,
            self.stepresolution_low,
        )
        self.arr_ktc_high = np.arange(
            self.ktc_high_min,
            self.ktc_high_max + self.stepresolution_high,
            self.stepresolution_high,
        )


def _make_template_df(arr_ktc_low, arr_ktc_high):
    """Create a pandas.DataFrame with all combinations of ktc_low and ktc_high

    Parameters
    ----------
    arr_ktc_low: numpy.array
    arr_ktc_high: numpy.array

    Returns
    -------
    df_cal_template: pandas.DataFrame. The DataFrame contains following columns:

        - *ktc_high* (float): high transport capacity coefficient
        - *ktc_low* (float); low transport capacity coefficient

    """
    df_cal_template = pd.DataFrame()
    x = np.unique(arr_ktc_low)
    y = np.unique(arr_ktc_high)
    arr_x, arr_y = np.meshgrid(x, y)
    df_cal_template["ktc_high"] = arr_x.flatten()
    df_cal_template["ktc_low"] = arr_y.flatten()

    return df_cal_template


def process_calibrationrun_output(
    txt_calibration,
    observed_sy,
    catchment_name,
    catchment_area,
    endpoint_coefficient=0.0,
):
    """Processes the output-file of a calibration run for a given catchment

    This function reads the calibration output file for a given catchement and
    calculates for every combination of ktc values the predicted specific
    sediment yield (P ssy) and calculates the squared difference between the
    observed and predicted (specific) sediment yield.

    Parameters
    ----------
    txt_calibration: str or pathlib.Path
        Path to the calibration-output file of a WaTEM/SEDEM model run, see
        :ref:`here <watemsedem:calibrationtxt>`.
    observed_sy: float
        The observed sediment yield for the catchment (in ton).
    catchment_name: str
        The name of the catchment.
    catchment_area: float
        The area of the catchment in ha.
    endpoint_coefficient: float, default 0.
        This coefficient determines the percentual amount of the sediment load
        that is captured in endpoints (sewers and ditches).
        In percentage (% - e.g. 80%). See also
        :func:`pywatemsedem.calibrate.check_endpoint_throughput_coefficient`.

    Returns
    -------
    df_cal: pandas.DataFrame
        DataFrame with all modelled sediment yields for different combinations
        of ktc-values. The DataFrame contains following columns:

        - *name* (string): name of the catchment.
        - *O_sy* (float): observed sediment yield (in ton).
        - *O_ssy* (float): observed specific sediment yield (ton/m$^2$).
        - *P_sy* (float): predicted sediment yield (in ton).
        - *P_ssy* (float): predicted specific sediment yield (ton/m$^2$).
        - *(O_sy-P_sy)* (float): difference between observed and predicted sy.
        - *|O_sy-P_sy|* (float): absolute difference between observed and predicted sy.
        - *(O_sy-P_sy)²* (float): squared difference between observed and predicted sy.
        - *(O_ssy-P_ssy)* (float): difference between observed and predicted ssy.
        - *|O_ssy-P_ssy|* (float): absolute difference between observed and predicted
          ssy.
        - *(O_ssy-P_ssy)²* (float): squared difference between observed and predicted
          ssy.
        - *ktc_low* (float): ktc low value used in the model run.
        - *ktc_high* (float): ktc high value used in the model run.

    """
    check_endpoint_throughput_coefficient(endpoint_coefficient)

    df_cal = pd.read_csv(txt_calibration, sep=";")
    df_cal.drop_duplicates(subset=["ktc_low", "ktc_high"], inplace=True)
    df_cal_template = _make_template_df(df_cal["ktc_low"], df_cal["ktc_high"])

    df_cal = pd.merge(df_cal, df_cal_template, how="right", on=["ktc_low", "ktc_high"])
    df_cal["name"] = catchment_name
    df_cal["O_sy"] = observed_sy
    df_cal["O_ssy"] = observed_sy / catchment_area
    df_cal["P_sy"] = (df_cal["outlet_1"]) / 1000

    if endpoint_coefficient != 0:
        if "sed_sewerin" in df_cal.columns:
            df_cal["P_sy"] = (
                df_cal["P_sy"]
                + df_cal["sed_sewerin"] / 1000 * endpoint_coefficient / 100
            )
        else:
            msg = (
                "No endpoints (sewerin-file) included in pywatemsedem model run, can not add"
                " percentage to estimated sediment yield (P_sy)."
            )
            logger.warning(msg)
    # convert kg to tons
    df_cal["P_ssy"] = df_cal["P_sy"] / catchment_area
    df_cal = compute_model_errors_sy(df_cal, "sy")
    df_cal = compute_model_errors_sy(df_cal, "ssy")

    return df_cal


def compute_model_errors_sy(df_cal, variable):
    """Compute model errors based on simulations and observations.

    - *Errors*: observation-prediction
    - *Absolute errors*: `|observation-prediction|`
    - *Squared errors*: `(observation-prediction)$^2$`

    Parameters
    ----------
    df_cal: pandas.DataFrame
        Definition, see :func:`pywatemsedem.calibrate.process_calibrationrun_output`.
    variable: str
        Name of output variable 'sy' or 'ssy'.

    Returns
    -------
    df_cal: pandas.DataFrame
        Updated with errors between simulation and observations.
    """
    if variable not in ["sy", "ssy"]:
        msg = (
            "Only model error based on sediment yield 'sy' and specific sediment"
            " yield 'ssy' can be computed."
        )
        raise IOError(msg)

    tag = f"O_{variable}-P_{variable}"
    df_cal[f"({tag})"] = df_cal[f"O_{variable}"] - df_cal[f"P_{variable}"]
    df_cal[f"|{tag}|"] = np.abs(df_cal[f"({tag})"])
    df_cal[f"({tag})²"] = df_cal[f"({tag})"] ** 2
    df_cal[f"({tag})²"] = np.where(
        df_cal["ktc_low"] > df_cal["ktc_high"], np.nan, df_cal[f"({tag})²"]
    )

    return df_cal


def check_endpoint_throughput_coefficient(endpoint_coefficient: float):
    """Check the endpoint (sewers and ditches) throughput coefficient used in
    :func:`pywatemsedem.calibrate.process_calibrationrun_output`.

    Raise error if the throughput coefficient is not within the defined ranges.

    Parameters
    ----------
    endpoint_coefficient: float
        Percentage of sediment load to captured in the endpoints (sewers and ditches)
        to consider for the computed predicted sediment load (see also column non river
        in scenario_x/modeloutput/calibration.txt).
    """
    cond = 0 <= endpoint_coefficient <= 100
    if not cond:
        msg = "Defined sewer throughput coefficient of"
        f"`{endpoint_coefficient}` for does not fall within allowed range [0,100]"
        raise IOError(msg)
    else:
        return endpoint_coefficient


def merge_calibration_results(lst_df_calibration, caldata):
    """Check if dimension of calibration space are equal to the one defined in
    the caldata instance (see :class:`pywatemsedem.calibrate.Calibration`)

    Merges a list of pandas.DataFrame with processed calibration model output

    Parameters
    ----------
    lst_df_calibration: list
        Every item in the list is a pandas.DataFrame. For the definition of the
        dataframe, see :func:`pywatemsedem.calibrate.process_calibrationrun_output`.
    caldata: pywatemsedem.core.calibrate.Calibration
        See :class:`pywatemsedem.calibrate.Calibration`

    Returns
    -------
    df_calibration_results: pandas.DataFrame
        DataFrame with all modelled sediment yields for different combinations
        of ktc-values. The DataFrame contains following columns:

        - *name* (string): name of the catchment.
        - *O_sy* (float): observed sediment yield.
        - *O_ssy* (float): observed specific sediment yield.
        - *P_sy* (float): predicted sediment yield.
        - *P_ssy* (float): predicted specific sediment yield.
        - *(O_sy-P_sy)²* (float): squared difference between observed and predicted sy.
        - *(O_ssy-P_ssy)²* (float): squared difference between observed and predicted
          ssy.
        - *ktc_low* (float): ktc low value used in the model run.
        - *ktc_high* (float): ktc high value used in the model run.
    """

    for df in lst_df_calibration:

        name = df["name"].iloc[0]
        msg = (
            f"Not all required calibration sets in the calibration are simulated"
            f" for `{name}`, please check you `calibration.txt` file in"
            f" scenario_x/modeloutput."
        )

        arr_ktc_low_exp = df["ktc_low"].values.astype(float)
        arr_ktc_high_exp = df["ktc_high"].values.astype(float)
        arr_ktc_low = np.arange(
            caldata.ktc_low_min, caldata.ktc_low_max, caldata.steps, dtype=float
        )
        arr_ktc_high = np.arange(
            caldata.ktc_high_min, caldata.ktc_high_max, caldata.steps, dtype=float
        )
        if np.array_equal(np.sort(arr_ktc_low), np.sort(arr_ktc_low_exp)):
            raise ValueError(msg)
        if np.array_equal(np.sort(arr_ktc_high), np.sort(arr_ktc_high_exp)):
            raise ValueError(msg)

    df_calibration_results = pd.concat(lst_df_calibration)

    return df_calibration_results


def linear_regression_evaluation_calibration(arr_observed, arr_predicted):
    """Compute linear regression between observed and predicted values.

    Parameters
    ----------
    arr_observed: np.ndarray or pandas.Series
        Observed (specific) sediment yield.
    arr_predicted: np.ndarray or pandas.Series
        Predicted (specific) sediment yield

    Returns
    -------
    p_reg: float
        See :class:`sklearn.linear_model.LinearRegression`
    r2: float
        See :func:`sklearn.metrics.r2_score`
    rico: float
        First order regression coefficient. See also
        :class:`sklearn.linear_model.LinearRegression`
    """
    regr = linear_model.LinearRegression(fit_intercept=False)
    arr_observed = arr_observed.values.reshape(-1, 1)
    arr_predicted = arr_predicted.values.reshape(-1, 1)
    regr.fit(arr_observed, arr_predicted)
    p_regr = regr.predict(arr_observed)
    r2 = r2_score(arr_predicted, p_regr)
    rico = regr.coef_[0][0]
    return p_regr, r2, rico


def calculate_model_efficiency(df_calibration_data, df_calibration_results):
    """Calculates the model efficiency (ME) for a calibration data set.

    The model efficiency calculations are for now based on the Nash-Sutcliff efficiency
    criterion (NSE).

    Parameters
    ----------
    df_calibration_data: pandas.DataFrame
        DataFrame with all necessary data for a calibration. It contains at
        least the following columns:

        - *name* (str): the name of a catchment.
        - *sy* (float): the sediment yield (ton).
        - *area* (float): the area of the catchment (ha).
        - *ssy* (float): the specific sediment yield (ton/ha).

    df_calibration_results: pandas.DataFrame
        DataFrame with all modelled sediment yields for different combinations
        of ktc-values and different catchments. For the definition of the
        dataframe, see :func:`pywatemsedem.calibrate.merge_calibration_results`.

    Returns
    -------
    df_me: pandas.DataFrame
        DataFrame with calculated model efficiencies for all different
        combinations of ktc-values. Contains following
        columns:

        - *ktc_low* (float): ktc low value used in the model run.
        - *ktc_high* (float): ktc high value used in the model run.
        - *ME sy* (float): model efficiency computed based on sediment yield.
        - *ME ssy* (float): model efficiency computed based on specific sediment yield.

    Notes
    -----
    Model efficiencies are computed based on observed and simulated yearly totals.

    """

    denominator_sy = compute_denomitor_nash_sutcliffe(
        df_calibration_data, "sy", squared_error=True
    )
    denominator_ssy = compute_denomitor_nash_sutcliffe(
        df_calibration_data, "ssy", squared_error=True
    )

    df_me_sy = compute_nash_sutcliffe(
        df_calibration_results, denominator_sy, "sy", squared_error=True
    )
    df_me_ssy = compute_nash_sutcliffe(
        df_calibration_results, denominator_ssy, "ssy", squared_error=True
    )

    df_me = df_me_sy.merge(df_me_ssy, on=["ktc_low", "ktc_high"])

    df_me = df_me.rename(columns={"NS sy": "ME sy", "NS ssy": "ME ssy"})

    return df_me


def compute_nash_sutcliffe(
    df_calibration_results, denominator, variable, squared_error=True
):
    """Compute Nash-Sutcliffe model efficiency for calibration results of WaTEM/SEDEM for a
    variable

    Parameters
    ----------
    df_calibration_results: pandas.DataFrame
        See :func:`pywatemsedem.calibrate.calculate_model_efficiency`.
    denominator: float
        See :func:`pywatemsedem.calibrate.compute_denomitor_nash_sutcliffe`.
    variable: str
        Name of output variable 'sy' or 'ssy'.
    squared_error: bool, default True
        Make use of squared errors to compute Nash-Sutcliffe (True/False).

    Returns
    -------
    df_me: pandas.DataFrame
        DataFrame with calculated model efficiencies for all different
        combinations of ktc-values. Contains following columns:

        - *ktc_low* (float): ktc low value used in the model run.
        - *ktc_high* (float): ktc high value used in the model run.
        - *NS sy* (float): Nash-Sutcliffe model efficiency computed based on sediment
          yield.
        - *NS ssy* (float): Nash-Sutcliffe model efficiency computed based on specific
          sediment yield.
    """
    if variable not in ["sy", "ssy"]:
        msg = (
            "Only model error based on sediment yield 'sy' and specific sediment"
            " yield 'ssy' can be computed."
        )
        raise IOError(msg)

    if squared_error:
        col = f"(O_{variable}-P_{variable})²"
    else:
        col = f"|O_{variable}-P_{variable}|"
    df_me = df_calibration_results.groupby(["ktc_low", "ktc_high"]).aggregate(
        {col: np.sum}
    )
    df_me[f"NS {variable}"] = 1 - (df_me[col] / denominator)
    df_me = df_me.reset_index()
    df_me[f"NS {variable}"] = np.where(
        df_me["ktc_high"] > df_me["ktc_low"], df_me[f"NS {variable}"], np.NaN
    )

    return df_me


def compute_denomitor_nash_sutcliffe(df_calibration_data, variable, squared_error):
    """Compute the denomitor for the Nash-Sutcliffe model efficiency.

    You can make use of squared or absolute errors to compute the denominator.

    Parameters
    ----------
    df_calibration_data: pandas.DataFrame
        See :func:`pywatemsedem.calibrate.calculate_model_efficiency`.
    variable: str, default 'sy'
        Name of output variable 'sy' or 'ssy'.
    squared_error: bool
        Make use of squared errors to compute denominator (True/False).

    Returns
    -------
    denominator: float
        Denominator of the Nash-Sutcliffe criterion.
    """

    if variable not in ["sy", "ssy"]:
        msg = (
            "Only model error based on sediment yield 'sy' and specific sediment"
            " yield 'ssy' can be computed."
        )
        raise IOError(msg)

    omean = df_calibration_data[variable].mean()
    df_calibration_data[f"(O_{variable} - O_{variable}_mean)"] = (
        df_calibration_data[f"{variable}"] - omean
    )

    if squared_error:
        denominator = (
            df_calibration_data[f"(O_{variable} - O_{variable}_mean)"] ** 2
        ).sum()
    else:
        denominator = (
            np.abs(df_calibration_data[f"(O_{variable} - O_{variable}_mean)"])
        ).sum()

    return denominator


def plot_model_efficiency(
    df_me, caldata, sy=True, contours=[-0.15, 0, 0.15, 0.3, 0.45, 0.6, 0.75]
):
    """This function plots the calculated model efficiency as a function of
    ktc-low and ktc-high.

    Parameters
    ----------
    df_me: pandas.DataFrame
        Dataframe with the calculated model efficiencies for all combinations
        of ktc-values. See :func:`pywatemsedem.calibrate.calculate_modelefficiency`
        for the defintion of the dataframe.
    caldata: pywatemsedem.core.calibrate.Calibration
        See :class:`pywatemsedem.calibrate.Calibration`.
    sy: bool, default True
        Plot the sediment yield (sy=True) or the specific sediment yield (sy=False).
    contours: list, default [-0.15, 0, 0.15, 0.3, 0.45, 0.6, 0.75]
        List with values for contour lines.

    Returns
    -------
    fig:  matplotlib.pyplot.figure
    ax:  matplotlib.pyplot.axes
    """

    field = "ME sy" if sy else "ME ssy"

    arr_me = np.reshape(
        df_me[field].values, (len(caldata.arr_ktc_low), len(caldata.arr_ktc_high))
    )
    extent = (
        caldata.ktc_high_min,
        caldata.ktc_high_max,
        caldata.ktc_low_min,
        caldata.ktc_low_max,
    )

    fig, ax = plt.subplots()

    norm = colors.Normalize(vmin=-1, vmax=1)
    cax = ax.imshow(
        arr_me, origin="lower", interpolation="none", extent=extent, cmap=cm.coolwarm
    )
    cax.set_norm(norm)

    arr_x, arr_y = np.meshgrid(caldata.arr_ktc_low, caldata.arr_ktc_high)
    CS = ax.contour(arr_x, arr_y, arr_me, contours, colors="k")

    ax.clabel(CS, inline=1, fontsize=10)

    # ax.plot([0, 1, 40], [0, 0.3, 12], color='k')
    # ax.plot([0, 1, 40], [0, 0.25, 10], color='b')
    # ax.plot([0, 1, 40], [0, 0.35, 14], color='b')

    xlabels = np.arange(
        caldata.ktc_high_min,
        caldata.ktc_high_max + caldata.stepresolution_high,
        round(caldata.steps / 4),
    )
    ylabels = np.arange(
        caldata.ktc_low_min,
        caldata.ktc_low_max + caldata.stepresolution_low,
        round(caldata.steps / 4),
    )
    plt.xticks(xlabels)
    plt.yticks(ylabels)
    plt.xlabel("ktc high (1/m)")
    plt.ylabel("ktc low (1/m)")
    fig.colorbar(cax)
    return fig, ax


def calculate_regressions(df_calibration_results, calibration):
    """Compute regression for plots

    Parameters
    ----------
    df_calibration_results: pandas.DataFrame
        DataFrame with all modelled sediment yields for different combinations
        of ktc-values and different catchments. For the definition of the
        dataframe, see :func:`pywatemsedem.calibrate.merge_calibration_results`.

    calibration: pywatemsedem.core.calibrate.Calibration
        See :class:`pywatemsedem.calibrate.Calibration`.

    Returns
    -------
    df_regressions: pandas.DataFrame
        DataFrame with the regression parameters for all combinations of ktc-values.
        The DataFrame contains following columns:

        - *ktc_low* (float): ktc low value used in the model run.
        - *ktc_high* (float): ktc high value used in the model run.
        - *R2 sy* (float): see
          :func:`pywatemsedem.calibrate.linear_regression_evaluation_calibration`.
        - *R2 ssy* (float): see
          :func:`pywatemsedem.calibrate.linear_regression_evaluation_calibration`.
        - *rico_sy* (float): see
          :func:`pywatemsedem.calibrate.linear_regression_evaluation_calibration`.
        - *rico_ssy* (float): see
          :func:`pywatemsedem.calibrate.linear_regression_evaluation_calibration`.

    """

    df_regressions = calibration.df_template.copy()
    df_regressions["R2_sy"] = np.nan
    df_regressions["R2_ssy"] = np.nan
    df_regressions["rico_sy"] = np.nan
    df_regressions["rico_ssy"] = np.nan

    for ktc_low in calibration.arr_ktc_low:
        for ktc_high in calibration.arr_ktc_high:
            sel = (df_regressions["ktc_low"] == ktc_low) & (
                df_regressions["ktc_high"] == ktc_high
            )

            if ktc_high >= ktc_low:
                selection = df_calibration_results.loc[
                    (df_calibration_results["ktc_low"] == ktc_low)
                    & (df_calibration_results["ktc_high"] == ktc_high)
                ]
                p_regr_sy, r2_sy, rico_sy = linear_regression_evaluation_calibration(
                    selection["O_sy"], selection["P_sy"]
                )
                p_regr_ssy, r2_ssy, rico_ssy = linear_regression_evaluation_calibration(
                    selection["O_ssy"], selection["P_ssy"]
                )

                df_regressions.loc[sel, "R2_sy"] = r2_sy
                df_regressions.loc[sel, "R2_ssy"] = r2_ssy
                df_regressions.loc[sel, "rico_sy"] = rico_sy
                df_regressions.loc[sel, "rico_ssy"] = rico_ssy
    return df_regressions


def calculate_ratio_ktc(df_ktc):
    """Calculates the ratio between ktc low and ktc high

    Parameters
    ----------
    df_ktc: pandas.DataFrame
        DataFrame with different comibinations of ktc-values. The DataFrame
        contains at least following columns:

        - *ktc_low* (float): ktc low value used in the model run.
        - *ktc_high* (float): ktc high value used in the model run.

    Returns
    -------
    df_ktc: pandas.DataFrame
        Identical DataFrame as the input DataFrame, but with an extra column:
        ratio_ktc (float).

    """
    df_ktc["ratio_ktc"] = df_ktc["ktc_low"] / df_ktc["ktc_high"]
    return df_ktc


def plot_regression(
    df, ktc_low, ktc_high, kind="sy", label_outliers=False, labels=False
):
    """Generate regression plot for calibration

    Parameters
    ----------
    df: pandas.DataFrame
        See :func:`pywatemsedem.calibrate.calculate_model_efficiency`.
    ktc_low: float
        Low transport capacity coefficient.
    ktc_high: float
        High transport capacity coefficient.
    kind: str
        "ssy" or "sy".
    label_outliers: bool, default False
        Flag if you want to label outliers in regression plot.
    labels: bool, default False
        Flag if you want to label all datapoints in regression plot.

    Returns
    -------
    fig: matplotlib.pyplot.figure
    ax1: matplotlib.pyplot.figure
    """
    selection = df.loc[(df["ktc_low"] == ktc_low) & (df["ktc_high"] == ktc_high)]
    if kind not in ["sy", "ssy"]:
        msg = f"kind `{kind}` is not a known pywatemsedem calibration plot type."
        raise IOError(msg)
    if kind == "sy":
        observed = selection["O_sy"]
        predicted = selection["P_sy"]
        title = "observed vs. predicted sediment yield"
    elif kind == "ssy":
        observed = selection["O_ssy"]
        predicted = selection["P_ssy"]
        title = "observed vs. predicted specific sediment yield"

    p_regr, r2, rico = linear_regression_evaluation_calibration(observed, predicted)

    fig = plt.figure()
    ax1 = fig.add_subplot(111)
    ax1.set_title(f"ktc-low:{ktc_low}, ktc-high:{ktc_high}\n{title}")

    if kind == "sy":
        unit = "ton"
    elif kind == "ssy":
        unit = "ton/ha"

    ax1.set_xlabel(f"Observed ({unit})")
    ax1.set_ylabel(f"Modelled ({unit})")
    ax1.scatter(x=observed, y=predicted, marker="o", s=13)
    ax_max = np.max([np.max(observed), np.max(predicted)])
    ax1.plot(
        [0, ax_max * 1.1], [0, ax_max * 1.1], color="grey", linewidth=0.75, ls="--"
    )
    ax1.plot(observed, p_regr, color="black", linewidth=0.75, label="Regression")

    if label_outliers:
        ax1 = add_outlier_names_to_calibration_plot(selection, ax1, kind)

    if labels:
        for i, txt in enumerate(selection["name"]):
            xpos = observed.iloc[i] + ax_max / 100
            ypos = predicted.iloc[i] + ax_max / 100
            ax1.annotate(txt, (xpos, ypos))

    txt = f"y = {round(rico, 2)}\n"
    txt += f"R2 ={round(r2_score(predicted, p_regr), 2)}"
    ax1.text(
        0.98,
        0.02,
        txt,
        horizontalalignment="right",
        verticalalignment="bottom",
        transform=ax1.transAxes,
    )
    ax1.set_xlim([0, ax_max * 1.1])
    ax1.set_ylim([0, ax_max * 1.1])
    plt.show()

    return fig, ax1


def add_outlier_names_to_calibration_plot(df, ax, kind, lq=50, uq=150):
    """Add `outlier id` or `catchment name` to outliers in evaluation plots.

    Outlier names are plotted when the quotient of observed to predicted is lower or
    higher than respectively the lower or upper quotient (lq and uq). For description
    of plots see :func:`pywatemsedem.calibrate.plot_regression`.

    Parameters
    ----------
    df: pandas.DataFrame

        Results of one set of ktc values with columns:

        - *O_sy* (float): observed (specific) sediment yield (ton).
        - *P_sy* (float): predicted (specific) sediment yield (ton/ha).

    ax: matplotlib.pyplot.axis.
    kind: str
        "ssy" or "sy".
    lq: str
        Lower quotient limit.
    uq: str
        Upper quotient limit.

    Returns
    -------
    ax: matplotlib.pyplot.axis.
    """
    df = df.reset_index()

    for index in df.index:
        x = df.loc[index, f"O_{kind}"]
        y = df.loc[index, f"P_{kind}"]
        col = df.loc[index, "name"] if "name" in df.columns else index
        quotient = x / y * 100
        if (quotient > uq) or (quotient < lq):
            ax.text(x, y, col)

    return ax



import logging
import tempfile
import warnings
from functools import wraps
from pathlib import Path

import fiona
import geopandas as gpd
import numpy as np
import pandas as pd
from matplotlib import colors
from matplotlib import pyplot as plt

from pywatemsedem.defaults import SAGA_FLAGS
from pywatemsedem.geo.factory import Factory
from pywatemsedem.geo.rasterproperties import RasterProperties
from pywatemsedem.geo.rasters import RasterMemory
from pywatemsedem.geo.utils import (
    any_equal_element_in_vector,
    clean_up_tempfiles,
    create_spatial_index,
    define_extent_from_vct,
    execute_subprocess,
    load_raster,
    read_rasterio_profile,
)
from pywatemsedem.io.folders import CatchmentFolder
from pywatemsedem.io.modeloutput import check_segment_edges
from pywatemsedem.tools import (
    format_forced_routing,
    get_df_area_unique_values_array,
    zip_folder,
)

from .errors import attribute_discrete_value_error, raster_discrete_value_error
from .valid import valid_req_property

# from .valid import valid_req_property
logger = logging.getLogger(__name__)


def valid_dtm(func):
    """Check if you have defined a DTM."""

    @wraps(func)
    def wrapper(self, *args, **kwargs):
        if self._dtm is None:
            msg = "Please first define a DTM!"
            raise IOError(msg)

        return func(self, *args, **kwargs)

    return wrapper


def valid_vct_river(func):
    """Check if you have defined a river vector."""

    @wraps(func)
    def wrapper(self, *args, **kwargs):
        if self._vct_river is None:
            msg = "Please define river vector!"
            raise IOError(msg)
        return func(self, *args, **kwargs)

    return wrapper


def valid_vct_infra_line(func):
    """Check if infrastructure vectors are defined"""

    @wraps(func)
    def wrapper(self, *args, **kwargs):
        if self._vct_infrastructure_roads is None:
            msg = "Please define infrastructure line vector!"
            raise IOError(msg)
        return func(self, *args, **kwargs)

    return wrapper


def valid_vct_infra_poly(func):
    """Check if infrastructure vectors are defined"""

    @wraps(func)
    def wrapper(self, *args, **kwargs):
        if self._vct_infrastructure_buildings is None:
            msg = "Please define infrastructure polygon vector!"
            raise IOError(msg)
        return func(self, *args, **kwargs)

    return wrapper


def valid_vct_parcels(func):
    """Check if infrastructure vectors are defined"""

    @wraps(func)
    def wrapper(self, *args, **kwargs):
        if self._vct_parcels is None:
            msg = "Please define parcels polygon vector!"
            raise IOError(msg)
        return func(self, *args, **kwargs)

    return wrapper


class Catchment(Factory):
    """Construct a new Catchment instance

    The catchment class holds all static information for a catchment.
    The class considers all data of which the content is user-option-independent (i.e.
    :class:`pywatemsedem.userchoices.UserChoices`.

    Following properties can be set:

        - *kfactor*: K-factor raster.
        - *landuse*: base landuse raster.
        - *river*: river line vector.
        - *water*: water polygon vector
        - *vct_infrastructure_buildings*: polygon vector of infrastructure
          (typically buildings).
        - *vct_infrastructure_roads*: roads line vector (with optional attribute
          'paved').
        - *cnsoil*: cn soil raster (CN-only).

    Notes
    -----
    Initialisation of mask, vector and raster handling is done with the help of the
    pywatemsedem factory class, see :class:`pywatemsedem.geo.factory.Factory`.
    """

    def __init__(
        self,
        name,
        vct_catchment,
        rst_dtm,
        resolution,
        epsg_code,
        nodata,
        results_folder=None,
    ):
        """Initialize

        Parameters
        ----------
        name: str
            Name of the catchment.
        vct_catchment: str or pathlib.Path
            Vector file of catchment outline (mask). This should be a single polygon
            vector.
        resolution: int
            Spatial resolution (m)
        epsg_code: int
            EPSG-code
        nodata: float
            Nodata-value
        results_folder: str | pathlib.Path, default None
            Folder path to write results to. If None, write to current folder
        """
        # prepare catchment
        if results_folder is None:
            results_folder = Path.cwd()
            msg = f"Setting results folder to {(results_folder / name)}"
            warnings.warn(msg)
        self.folder = CatchmentFolder(Path(results_folder) / name, resolution)
        self.folder.create_all()
        self.name = name

        # initiate factory
        super().__init__(resolution, epsg_code, nodata, results_folder / name)

        # if the binary mask of the catchment is a geopandas dataframe, map it first
        # to a file
        if type(vct_catchment) is gpd.GeoDataFrame:
            self.vct_catchment = self.folder.vct_folder / "catchment.shp"
            vct_catchment.to_file(self.vct_catchment)
        else:
            self.vct_catchment = vct_catchment
        # input your own raster properties and do not self-generate them.
        self.create_rasterproperties = False

        # get RasterProperties from catchment vector and bounds DTM
        rp = read_rasterio_profile(rst_dtm)
        bounds = RasterProperties.from_rasterio(rp).bounds
        self.rp = define_extent_from_vct(
            self.vct_catchment,
            resolution,
            nodata,
            epsg_code,
            bounds=bounds,
        )
        # set mask
        self.mask = self.vct_catchment

        # set dtm
        self.dtm = rst_dtm

        # attributes
        self._pfactor = None
        self._kfactor = None
        self._segments = None
        self._routing = None
        self._hydrosoilgroup = None
        self._landuse = None
        self._water = None
        self._infrastructure = None

        # set name and logger
        self.name = str(name)

        # set other attributes to none
        self._river = None
        self._vct_river = None
        self._vct_tubed_river = None
        self._vct_water = None
        self._vct_infrastructure_buildings = None
        self._vct_infrastructure_roads = None
        self._adjacent_edges = None
        self._up_edges = None

    def zip(self):
        """Zip catchment folder"""
        zip_folder((self.folder / "Data_Bekken"))

    @property
    def dtm(self):
        """Getter dtm raster

        Returns
        -------
        pywatemsedem.geo.rasters.AbstractRaster
            Float64 raster with DTM height values.
        """
        return self._dtm

    @dtm.setter
    def dtm(self, raster_input):
        """Assign digital elevation model raster

        The DTM input should contain valid DTM raster values in meter a.s.l.

        Parameters
        ----------
        raster_input: Pathlib.Path, str or numpy.array
            File path/array digital terrain model (m).
        """
        self._dtm = self.raster_factory(raster_input, flag_mask=False, flag_clip=True)

        def filter_within_parcels():
            """Filter dtm within parcel boundaries"""
            temp_landuse = Path(tempfile.NamedTemporaryFile(suffix=".rst").name)
            temp_dtm_in = Path(tempfile.NamedTemporaryFile(suffix=".rst").name)
            temp_dtm = Path(tempfile.NamedTemporaryFile(suffix=".sgrd").name)
            self.dtm.write(temp_dtm_in)
            self.parcels.write(temp_landuse)

            cmd_args = ["saga_cmd", "-f=s", "watem", "2"]
            cmd_args += ["-DEM", str(temp_dtm_in)]
            cmd_args += ["-PRC", str(temp_landuse)]
            cmd_args += ["-DEM_FILTER", str(temp_dtm)]
            msg = "failed to filter_within_parcels dtm"
            execute_subprocess(cmd_args, msg)
            self._dtm._arr, _ = load_raster(temp_dtm.with_suffix(".sdat"))
            clean_up_tempfiles(temp_landuse, "tiff")
            clean_up_tempfiles(temp_dtm, "saga")
            clean_up_tempfiles(temp_dtm_in, "rst")

        self._dtm.filter = filter_within_parcels

    @property
    def pfactor(self):
        """P-factor raster

        The P-factor raster holds values between 0 and 1, and is the factor in the
        RUSLE equation which takes into account erosion control. In pywatemsedem, this
        factor is always set to 1, as erosion source-oriented control measures are
        accounted for in the C-factor.

        Returns
        -------
        pywatemsedem.geo.rasters.AbstractRaster
        """
        return self._mask

    @property
    def kfactor(self):
        """Getter K-factor raster

        The K-factor raster is a raster holding K-values for every pixel
        (:math:`\\frac{ton.h}{MJ.mm}`).

        Returns
        -------
        pywatemsedem.geo.rasters.AbstractRaster
        """
        return self._kfactor

    @kfactor.setter
    def kfactor(self, raster_input):
        """Assign K-factor raster

        The K-factor raster is a raster holding K-values for every pixel
        (:math:`\\frac{ton.h}{MJ.mm}`). Negative no nodata values are set to zero.

        Parameters
        ----------
        raster_input: Pathlib.Path, str or numpy.array
            File path/array K-factor raster (:math:`\\frac{ton.h}{MJ.mm}`) with K-values
            (see K-factor in RUSLE equation).

        Notes
        -----
        1. The input requires as unit: :math:`\\frac{ton.h}{MJ.mm}`, whereas the output
           unit is :math:`\\frac{kg.h}{MJ.mm}`.
        2. The input data is rounded to kg (no decimals) by a conversion to int32.
        """
        self._kfactor = self.raster_factory(
            raster_input, flag_mask=True, flag_clip=True
        )
        cond = self._kfactor.arr != self._kfactor.rp.nodata
        cond_neg = cond & (self._kfactor.arr < 0)
        if np.any(cond_neg):
            msg = (
                "Negative values detected in K-factor raster, setting negative "
                "values to 0."
            )
            warnings.warn(msg, UserWarning)
            self._kfactor.arr[cond_neg] = 0.0
        self._kfactor.arr[cond] = self._kfactor.arr[cond] * 1000
        self._kfactor.arr[cond] = np.round(self._kfactor.arr[cond])
        self._kfactor.arr = self._kfactor.arr.astype(np.int32)

    @property
    def hydrological_soil_group(self):
        """CN hydrological soil group raster

        This raster (int) holds the four categories (see [2]_):

        - *A* (value 1): Soils in this group have low runoff potential when thoroughly
          wet.
        - *B* (value 2): Soils in this group have moderately low runoff potential when
          thoroughly wet.
        - *C* (value 3): Soils in this group have moderately high runoff potential when
          thoroughly wet.
        - *D* (value 4): Soils in this group have high runoff potential when thoroughly
          wet.

        Returns
        -------
        pywatemsedem.geo.rasters.AbstractRaster
        """
        return self._hydrosoilgroup

    @hydrological_soil_group.setter
    # @valid_req_property(req_property_name="dtm", mandatory=True)
    def hydrological_soil_group(self, raster_input):
        """Assign CN soil raster

        This raster (int) holds the four categories (see [2]_):

        - *A* (value 1): Soils in this group have low runoff potential when thoroughly
          wet.
        - *B* (value 2): Soils in this group have moderately low runoff potential when
          thoroughly wet.
        - *C* (value 3): Soils in this group have moderately high runoff potential when
          thoroughly wet.
        - *D* (value 4): Soils in this group have high runoff potential when thoroughly
          wet.

        Parameters
        ----------
        raster_input: Pathlib.Path, str or numpy.array
            File path/array base landuse raster (class) with values *1*: A, *2*: B,
            *3*: C and *4*: D.

        References
        ----------
        .. [1] NRCS, 2004. Hydrologic soil-cover complexes (chapter 9).
        .. [2] NRCS, 2009. Hydrologic soil groups (chapter 7).
        """
        self._hydrosoilgroup = self.raster_factory(
            raster_input, flag_mask=False, flag_clip=True
        )
        raster_discrete_value_error(
            self._hydrosoilgroup.arr,
            "Hydrological soil group",
            [self._hydrosoilgroup.rp.nodata, 1, 2, 3, 4],
            [
                "nodata",
                "A: Soils with low runoff potential when thoroughly wet.",
                "B: Soils with moderately low runoff potential when thoroughly wet",
                "C: Soils with moderately high runoff potential when thoroughly wet",
                "D: Soils with high  runoff potential when thoroughly wet",
            ],
        )

    @property
    def landuse(self):
        """Getter background land-use raster

        Returns
        -------
        pywatemsedem.geo.rasters.AbstractRaster
            The raster can following values:

            - *nodata*: nodata
            - *-5*: open water
            - *-4*: grass
            - *-3*: forest
            - *-2*: infrastructure
            - *10*: agricultural fields
        """
        return self._landuse

    @landuse.setter
    # @valid_req_property(req_property_name="dtm", mandatory=True)
    def landuse(self, raster_input):
        """Assign background land-use raster

        Raster should contain:

            - *-9999*: nodata
            - *-5*: open water
            - *-4*: grass
            - *-3*: forest
            - *-2*: infrastructure
            - *10*: agricultural fields

        Rasters with values outside this range return an error.

        Parameters
        ----------
        raster_input: Pathlib.Path, str or numpy.array
            File path/array base landuse raster (class) with values *-9999*: nodata,
            *-5*: open water, *-4*: grass, *-3*: forest, *-2*: infrastructure, *10*:
            agricultural fields.

        """
        self._landuse = self.raster_factory(
            raster_input, flag_mask=True, flag_clip=True
        )
        raster_discrete_value_error(
            self._landuse.arr,
            "Landuse",
            [self._landuse.rp.nodata, -6, -5, -4, -3, -2, 10],
            [
                "nodata",
                "grass strips",
                "open water",
                "grass",
                " forest",
                "infrastructure",
                "agricultural fields",
            ],
        )

        def plot(nodata=None, *args, **kwargs):
            """Plotting fun"""
            plt.subplots(figsize=[10, 10])

            cmap = colors.ListedColormap(
                [
                    "#64cf1b",
                    "#3b7db4",
                    "#71b651",
                    "#387b00",
                    "#000000",
                    "#00bfff",
                    "#ffffff",
                    "#a47158",
                ]
            )
            bounds = [-6.5, -5.5, -4.5, -3.5, -2.5, -1.5, -0.5, 0.5, 1.5]
            norm = colors.BoundaryNorm(bounds, cmap.N)
            arr = self._landuse.arr.copy().astype(np.float32)
            if nodata is not None:
                arr[arr == nodata] = np.nan
            img = plt.imshow(arr, cmap=cmap, norm=norm, *args, **kwargs)
            cbar = plt.colorbar(
                img,
                cmap=cmap,
                norm=norm,
                boundaries=bounds,
                ticks=[-6, -5, -4, -3, -2, -1, 0, 1],
                shrink=0.5,
            )
            cbar.ax.set_yticklabels(
                [
                    "Grass strips (-6)",
                    "Pools (-5)",
                    "Meadow (-4)",
                    "Forest (-3)",
                    "Infrastructure (-2)",
                    "River (-1)",
                    "Outside boundaries (0)",
                    "Agriculture (>0)",
                ]
            )
            plt.show()

        self._landuse.plot = plot

    @staticmethod
    def create_height_contours(rst_input, vct_output):
        """Create height contours line vector based on a DTM raster

        Parameters
        ----------
        rst_input: pathlib.Path
            Input DTM raster
        vct_output: pathlib.Path
            Output height contour line vector
        """
        if rst_input is not None:
            if not vct_output.exists():
                cmd_args = ["saga_cmd", SAGA_FLAGS, "shapes_grid", "5"]
                cmd_args += [
                    "-GRID",
                    str(rst_input),
                    "-CONTOUR",
                    str(vct_output),
                    "-ZSTEP",
                    "1",
                ]
                try:
                    execute_subprocess(cmd_args)
                except OSError as e:
                    if "corrupted size vs. prev_size in fastbins" in str(e):
                        try:
                            fiona.open(vct_output)
                        except fiona.errors.DriverError:
                            raise IOError(e)
                create_spatial_index(vct_output)
        else:
            msg = "Inputfile is not valid!"
            raise FileNotFoundError(msg)

    @property
    def catchment_area(self):
        """Total area of the catchment

        Returns
        -------
        float: :math:`m^2`
        """
        return self.vct_catchment.geodata.area

    @property
    # @valid_req_property(req_property_name="vct_river", mandatory=True)
    def river_segment_length(self):
        """Total length of all river segments in the catchment

        Returns
        -------
        float, meter
        """
        return self._vct_river.geodata.length.sum()

    @property
    # @valid_req_property(req_property_name="vct_river", mandatory=True)
    def river(self):
        """Getter river raster

        Returns
        -------
        pywatemsedem.geo.rasters.AbstractRaster
            Raster containing following values:

            - *-1*: river
            - *0*: no river
            - *-9999*: nodata
        """
        arr = np.where(self._river.arr > 0, -1, self._river.rp.nodata)

        return RasterMemory(arr, self.rp)

    @property
    # @valid_req_property(req_property_name="vct_river", mandatory=False)
    def adjacent_edges(self):
        """Getter dataframe adjacent edges/segment property

        This property is generated by assigning a river line vector to the catchment
        class. Every row indicates a connection between two segments: segment *from*
        (column 1) flows into segment to (column 2). See :ref:`here <watemsedem:adjsegments>`

        Returns
        -------
        pandas.DataFrame
            With columns, flow

                - *from* (int): from segment
                - *to* (int): to segment
        """
        return self._adjacent_edges

    @property
    # @valid_req_property(req_property_name="vct_river", mandatory=False)
    def up_edges(self):
        """Getter dataframe upstream edges/segment property

        This property is generated by assigning a river line vector to the catchment
        class. See :ref:`here <watemsedem:upstrsegments>`

        Returns
        -------
        pandas.DataFrame
            With columns

                - *edge* (int): segment id
                - *upstream edge* (int): segment id of one of the upstream segment of
                  the edge.
                - *proportion* (float): proportion of upstream edge flowing into edge.
                  Lower than 1 when upstream edge flows into two edges.

        """
        return self._up_edges

    @property
    def vct_river(self):
        """Assign river-line vector

        The river vector should be a line-vector file. No specific attributes should be
        defined.

        Parameters
        ----------
        vector_input: Pathlib.Path, str or geopandas.GeoDataFrame
            Line vector file.
        """
        return self._vct_river

    @vct_river.setter
    def vct_river(self, vector_input):
        """Assign river-line vector

        The river vector should be a line-vector file. No specific attributes should be
        defined.

        Parameters
        ----------
        vector_input: Pathlib.Path, str or geopandas.GeoDataFrame
            Line vector file. With optional

                - *underground* (int): does river section run underground
        """
        # clip
        vct_river_clipped = self.folder.vct_folder / f"river_{self.name}.shp"
        # remove temporary files
        clean_up_tempfiles(vct_river_clipped, "shp")
        self._vct_river = self.vector_factory(
            vector_input, "LineString", allow_empty=True
        )
        self._vct_river.write(vct_river_clipped)

        # set topologize
        logger.info("Preparing river topology...")
        vct_river_topo = self.folder.vct_folder / f"topology_{self.name}.shp"
        clean_up_tempfiles(vct_river_topo, "shp")
        self._adjacent_edges, self._up_edges = self.topologize_river(
            vct_river_clipped, vct_river_topo, self.mask_raster
        )

        self._vct_river = self.vector_factory(
            vct_river_topo, "LineString", allow_empty=True
        )

        river = self._vct_river.rasterize(
            self.mask_raster, self.rp.epsg, "line_id", "integer", gdal=True
        )
        self._river = self.raster_factory(river, allow_nodata_array=True)
        self._adjacent_edges, self._up_edges, flag = check_segment_edges(
            self.adjacent_edges, self.up_edges, self._river.arr
        )

        # set routing
        if self._vct_river.geodata.shape[0] > 0:
            routing = self._vct_river.rasterize(
                self.mask_raster,
                self.rp.epsg,
                convert_lines_to_direction=True,
                gdal=True,
            )
            routing[routing == self.rp.nodata] = 0
            self._routing = self.raster_factory(routing, flag_mask=False)
        else:
            msg = "River input vector is empty, setting river routing to None"
            logger.info(msg)
            self._routing = None

    @property
    def vct_tubed_river(self):
        """Assign underground river-line vector

        The river tubed vector should be a line-vector file. No specific
        attributes should be defined.

        Parameters
        ----------
        vector_input: Pathlib.Path, str or geopandas.GeoDataFrame
            Line vector file.
        """
        return self._vct_tubed_river

    @vct_tubed_river.setter
    def vct_tubed_river(self, vector_input):
        """Assign underground river-line vector

        The river tubed vector should be a line-vector file. No specific
        attributes should be defined.

        Parameters
        ----------
        vector_input: Pathlib.Path, str or geopandas.GeoDataFrame
            Line vector file.
        """
        valid_req_property(
            self,
            current_property="vct_tubed_river",
            req_property_name="vct_river",
            mandatory=True,
        )

        msg = (
            "A vector is assigned to tubed rivers, please "
            "set 'Force Routing' in model option equal to 1."
        )
        warnings.warn(msg)
        vct_tubed_river = self.vector_factory(
            vector_input, "LineString", allow_empty=False
        )
        self._vct_tubed_river = format_forced_routing(
            vct_tubed_river.geodata,
            self.rp.gdal_profile["minmax"],
            self.rp.resolution,
        )
        if self.vct_river is not None:
            if any_equal_element_in_vector(
                self.vct_river.geodata.geometry, vct_tubed_river.geodata.geometry
            ):
                msg = (
                    "Lines in the river (tubed) vector are also present in the "
                    "river vector. Please remove overlapping vectors."
                )
                raise IOError(msg)

    @property
    def routing(self):
        """Getter river routing raster

        See :ref:`here <watemsedem:riverroutingmap>`

        Returns
        -------
        pywatemsedem.geo.rasters.AbstractRaster or None
            Raster containing values from 0 to 8:

            - *0*: do not route further
            - *1*: route to upper pixel
            - *2*: route to upper right pixel
            - *3*: route to right pixel
            - *4*: route to lower right pixel
            - *5*: route to lower pixel
            - *6*: route to lower left pixel
            - *7*: route to left pixel
            - *8*: route to upper left pixel
            - *-9999*: nodata
        """
        return self._routing

    @property
    # @valid_req_property(req_property_name="vct_river", mandatory=False)
    def segments(self):
        """Getter river segments raster

        See :ref:`here <watemsedem:riversegmentfile>`

        Returns
        -------
        pywatemsedem.geo.rasters.AbstractRaster
            Raster containing following values:

            - *>0*: segment_id
            - *-9999*: nodata

        Notes
        -----
        The number of segments is limited to int16.
        """
        self._segments = self.raster_factory(
            self._river.arr, self._river.rp, allow_nodata_array=True
        )
        self._segments.arr = np.where(self._segments.arr < 1, 0, self._segments.arr)
        self._segments.arr = self._segments.arr.astype(np.int16)

        return self._segments

    @staticmethod
    def topologize_river(
        vct_input,
        vct_output,
        rst_mask,
        tolerance=None,
    ):
        """Prepare topology of the river segments with SAGA-GIS

        Parameters
        ----------
        vct_input: str or pathlib.Path
            Line input vector
        vct_output: str of pathlib.Path
            Topologized line output vector
        rst_mask: str or pathlib.Path
            Mask raster (used for mapping)
        tolerance: float
            If not None, the ``TOLERANCE`` command line argument of the saga topology
            command is given this value.

        Returns
        -------
        adjacent_edges: pandas.DataFrame
           See :func:`pywatemsedem.catchment.Catchment.adjacent_edges` and
           :ref:`here <watemsedem:adjsegments>`.
        up_edges: pandas.DataFrame
            See :func:`pywatemsedem.catchment.Catchment.up_edges` and
            :ref:`here <watemsedem:upstrsegments>`.
        """
        cmd_args = ["saga_cmd", SAGA_FLAGS, "topology", "0"]
        cmd_args += ["-INPUTLINES", str(vct_input)]
        cmd_args += ["-OUTPUTLINES", str(vct_output)]
        cmd_args += ["-SYSTEM", str(rst_mask)]
        if tolerance is not None:
            cmd_args += ["-TOLERANCE", str(tolerance)]

        temp_adjacent_edges = tempfile.NamedTemporaryFile(suffix=".txt").name
        temp_up_edges = tempfile.NamedTemporaryFile(suffix=".txt").name

        # temp fix
        try:
            execute_subprocess(cmd_args)
        except OSError as e:
            if "corrupted size vs. prev_size in fastbins" in str(e):
                try:
                    fiona.open(vct_output)
                except fiona.errors.DriverError:
                    raise IOError(e)

        cmd_args = ["saga_cmd", SAGA_FLAGS, "topology", "1"]
        cmd_args += ["-INPUTLINES", str(vct_output)]
        cmd_args += ["-ADJECANT_EDGES", str(temp_adjacent_edges)]
        cmd_args += ["-UPSTREAM_EDGES", str(temp_up_edges)]

        # temp fix
        try:
            execute_subprocess(cmd_args)
        except OSError as e:
            if "corrupted size vs. prev_size in fastbins" in str(e):
                try:
                    pd.read_csv(str(temp_adjacent_edges), sep="\t")
                    pd.read_csv(str(temp_up_edges), sep="\t")
                except fiona.errors.DriverError:
                    raise IOError(e)

        adjacent_edges = pd.read_csv(str(temp_adjacent_edges), sep="\t")
        up_edges = pd.read_csv(str(temp_up_edges), sep="\t")

        clean_up_tempfiles(Path(temp_up_edges), "txt")
        clean_up_tempfiles(Path(temp_adjacent_edges), "txt")

        return adjacent_edges, up_edges

    @property
    def vct_water(self):
        """Getter water polygon vector

        Returns
        -------
        pywatemsedem.geo.vectors.AbstractVector
            Polygon vector with attribute "value"
        """

        return self._vct_water

    @vct_water.setter
    def vct_water(self, vector_input):
        """Assign water polygon vector

        The water vector should be a polygon-vector file. No specific attributes should
        be defined.

        Parameters
        ----------
        vector_input: Pathlib.Path, str or geopandas.GeoDataFrame
            Polygon vector file

        Notes
        -----
        Water is mapped to the value of -5 and considered as a sink for sediment in
        WaTEM/SEDEM.
        """
        self._vct_water = self.vector_factory(vector_input, "Polygon")
        self._vct_water.geodata["value"] = -5

    @property
    # @valid_req_property(req_property_name="vct_water", mandatory=False)
    def water(self):
        """Getter water polygon vector

        Returns
        -------
        pywatemsedem.geo.rasters.AbstractRaster
            Raster containing following values:

            - *-5*: open water
            - *-9999*: nodata
        """
        self._vct_water.geodata["value"] = self._vct_water.geodata["value"].astype(
            float
        )
        water = self._vct_water.rasterize(
            self.mask_raster, self.rp.epsg, "value", "integer", gdal=True
        )

        return self.raster_factory(water)

    @property
    def vct_infrastructure_buildings(self):
        """Getter infrastructure polygon vector

        Returns
        -------
        pywatemsedem.geo.vectors.AbstractVector
            Polygon vector with attribute "paved"
        """

        return self._vct_infrastructure_buildings

    @vct_infrastructure_buildings.setter
    def vct_infrastructure_buildings(self, vector_input):
        """Assign infrastructure polygon vector

        The infrastructure buildings vector should be polygon vector. Buildings are
        defined as all non-road elements.

        Parameters
        ----------
        vector_input: Pathlib.Path, str or geopandas.GeoDataFrame
            Polygon vector file
        """
        self._vct_infrastructure_buildings = self.vector_factory(
            vector_input, "Polygon"
        )
        self._vct_infrastructure_buildings.geodata["paved"] = -2

    @property
    # @valid_req_property(
    #    req_property_name="vct_infrastructure_buildings", mandatory=False
    # )
    def infrastructure_buildings(self):
        """Getter rasterized infrastructure polygon vector

        Returns
        -------
        pywatemsedem.geo.rasters.AbstractRaster
            Raster containing following values:

            - *-2*: paved
            - *-9999*: nodata
        """
        self._vct_infrastructure_buildings.geodata[
            "paved"
        ] = self._vct_infrastructure_buildings.geodata["paved"].astype(float)

        infra = self._vct_infrastructure_buildings.rasterize(
            self.mask_raster, self.rp.epsg, col="paved", gdal=True
        )

        return self.raster_factory(infra)

    @property
    # @valid_req_property(req_property_name="vct_infrastructure_roads", mandatory=False)
    def infrastructure_roads(self):
        """Getter rasterized infrastructure polygon vector

        Returns
        -------
        pywatemsedem.geo.rasters.AbstractRaster
            Raster containing following values:

            - *-2*: paved
            - *-7*: paved
            - *-9999*: nodata
        """
        self._vct_infrastructure_roads.geodata[
            "paved"
        ] = self._vct_infrastructure_roads.geodata["paved"].astype(float)

        arr = self._vct_infrastructure_roads.rasterize(
            self.mask_raster, self.rp.epsg, col="paved", gdal=True
        )
        return self.raster_factory(arr)

    @property
    def vct_infrastructure_roads(self):
        """Getter infrastructure line vector"""
        return self._vct_infrastructure_roads

    @vct_infrastructure_roads.setter
    def vct_infrastructure_roads(self, vector_input):
        """Assign infrastructure line vector

        The infrastructure roads vector should be polygon vector. Paved and non-paved
        roads can be defined:

            - *-2*: paved
            - *-7*: non-paved

        Parameters
        ----------
        vector_input: Pathlib.Path, str or geopandas.GeoDataFrame
            Line vector file with allowed (optional) column 'paved': allowed values -2
            (paved) and -7 (non-paved).
        """
        self._vct_infrastructure_roads = self.vector_factory(vector_input, "LineString")
        geodata = self._vct_infrastructure_roads.geodata
        if "paved" in self._vct_infrastructure_roads.geodata:
            attribute_discrete_value_error(
                geodata,
                "Infrastucture roads",
                "paved",
                {self.rp.nodata, -2, -7},
                {"nodata", "paved", " non-paved"},
            )
        else:
            geodata["paved"] = -2

        self._vct_infrastructure_roads.geodata = geodata

    @property
    def infrastructure(self):
        """Get infrastucture raster

        Get rasterized polygon and line vectors. The procedure adds the line data
        (roads) to the  polygon data (buildings). As such buildings are considered as
        the base map.

        Returns
        -------
        pywatemsedem.geo.rasters.AbstractRaster
            Raster containing following values:

            - *-2*: paved
            - *-7*: paved

        Notes
        -----
        Note that current module allows to either define a roads (1), buildings (2)
        or roads and buildings (3). If no roads or buildings are defined, an error is
        thrown."""

        if self.infrastructure_roads is not None:
            if self.infrastructure_buildings is not None:
                arr1 = self.infrastructure_roads.arr.copy()
                arr2 = self.infrastructure_buildings.arr.copy()
                cond1 = arr1 == self.infrastructure_roads.rp.nodata
                cond2 = arr2 != self.infrastructure_buildings.rp.nodata
                arr = np.where(cond1 & cond2, arr2, arr1)
                arr = np.where(
                    arr != self.infrastructure_buildings.rp.nodata, arr, self.rp.nodata
                )
                self._infrastructure = RasterMemory(arr, self.rp)
            else:
                self._infrastructure = self.infrastructure_roads
        else:
            if self.infrastructure_buildings is not None:
                self._infrastructure = self.infrastructure_buildings
            else:
                msg = (
                    "First define infrastructure lines (roads) or/and polygons "
                    "(buildings) before calling infrastructure property!"
                )
                raise IOError(msg)

        def plot(nodata=None, *args, **kwargs):
            """Plotting fun"""
            fig, ax = plt.subplots(figsize=[10, 10])
            """Plot infrastructure"""
            arr_plot = self._infrastructure.arr.copy().astype(np.float32)
            if nodata is not None:
                arr_plot = np.where(arr_plot != nodata, 1, arr_plot)
                arr_plot = np.where(arr_plot == nodata, 0, arr)

            im = ax.imshow(arr_plot, *args, **kwargs)
            fig.colorbar(im, ax=ax, shrink=0.5)

        self._infrastructure.plot = plot

        return self._infrastructure

    def _calculate_soil_statistics(self):
        """Calculate statistics of K-factor map

        Calculates the area (m²) and the relative amount of appearance of
        every K-factor in the catchment. The results are written to a csv-file
        "K_factor_stats_[resolution].csv" in ``self.cfolder.rst_folder ``.

        #TODO

        """
        fname = self.folder.rst_folder / f"K_factor_stats_{self.rp.resolution}.csv"
        if not fname.exists():
            msg = "Calculating K-factor statistics"
            logger.info(msg)
            df = get_df_area_unique_values_array(self.kfactor.arr, self.rp.resolution)
            df.drop(df[df["values"] == self.rp.nodata].index, inplace=True)
            total_area = df["area"].sum()
            df["rel_area"] = (df["area"] / total_area) * 100
            df.to_csv(fname, index=False)

    def create_folder_structure(self):
        """Creates a fixed folder structure for the model data

        The folder layout will have the following directory structure:

        ::

            ├── homefolder
            │    └── bekken folder / catchment folder
            │         ├── Rsts
            │         ├── Shps

        """
        logger.info("Aanmaken folderstructuur bekken...")
        self.folder = CatchmentFolder(Path(self.name), self.rp.resolution)
        self.folder.create_all()



from dataclasses import dataclass

import matplotlib as mpl
import numpy as np
import pandas as pd

# hvplot functionalities
from matplotlib import colors

from pywatemsedem.geo.factory import Factory
from pywatemsedem.geo.utils import (
    check_raster_properties_raster_with_template,
    mask_array_with_val,
)
from pywatemsedem.io.plots import (
    axes_creator,
    plot_continuous_raster,
    plot_discrete_raster,
)
from pywatemsedem.io.valid import (
    valid_array_type,
    valid_boundaries,
    valid_nodata,
    valid_non_nan,
    valid_values,
)

IMPLEMENTED_RASTER_TYPES = [np.int16, np.int32, np.int64, np.float32, np.float64]
COLORMAP = "cividis"


def valid_segments(func):
    """Check if you have defined a river segments raster."""

    def wrapper(self, *args, **kwargs):
        if self._riversegments is None:
            msg = (
                f"Please first define WaTEM/SEDEM river segments, see "
                f"{Modelinput.riversegments}!"
            )
            raise IOError(msg)
        return func(self, *args, **kwargs)

    return wrapper


@dataclass
class Modelinput(Factory):
    def __init__(self, template, resolution, epsg, nodata):
        """AbstractRaster class with model inputs as attributes. Modelinput class
        serves the goal of automating the reading in, checking and visualisation
        of the input data of the WaTEM/SEDEM model.

        Parameters
        ----------
        template: pathlib.Path | str
            Template raster needed to extract raster properties.
            Raster contains values in [0,1]. Values in ]0,1] define the catchment.
        resolution: int
            See :class:`pywatemsedem.geo.RasterProperties`
        epsg: int
            See :class:`pywatemsedem.geo.RasterProperties`.
        nodata: int
            See :class:`pywatemsedem.geo.RasterProperties`.
        """
        # apply factory and set mask
        super().__init__(resolution, epsg, nodata, template.parent)
        self.mask = template

        # DATA
        self._cfactor = None  # attribute
        self._buffers = None
        self._dtm = None
        self._kfactor = None
        self._ktc = None
        self._outlet = None
        self._pfactor = None
        self._compositelanduse = None
        self._ptef = None
        self._riversegments = None
        self._riverrouting = None
        self._sewers = None
        self._upstreamsegments = None
        self._adjacentsegments = None

    @property
    def cfactor(self):
        """Getter cfactor attribute.

        For documentation, see :ref:`here <watemsedem:cmap>`
        """
        return self._cfactor

    @cfactor.setter
    def cfactor(self, raster):
        """Setter

        Parameters
        ----------
        raster: pathlib.Path | str
        """
        self._cfactor = self.raster_factory(raster, flag_mask=False)
        # checks on raw raster data
        valid_non_nan(self.cfactor.arr)
        valid_array_type(self.cfactor.arr, required_type=np.float32)
        lower, upper = 0, 1
        valid_boundaries(
            self.cfactor.arr[self.cfactor.arr != self.rp.nodata],
            lower=lower,
            upper=upper,
        )
        check_raster_properties_raster_with_template(self.rp, raster, epsg=self.rp.epsg)

        def plot(fig=None, ax=None, *args, **kwargs):
            """Plot for _cfactor with cividis as colormap

            Parameters
            ----------
            fig: matplotlib.figure.Figure, default = None
                if not given, defaults to generating new figure
            ax: matplotlib.pyplot.axis, default = None
                if not given, defaults to generating new axis

            Returns
            -------
            fig: matplotlib.figure.Figure

            ax: matplotlib.axes.Axes
            """
            fig, ax = axes_creator(fig, ax)
            arr = mask_array_with_val(self.cfactor.arr, self.mask.arr, 0)
            fig, ax = plot_continuous_raster(
                fig, ax, arr, self.rp.bounds, vmin=lower, vmax=upper, *args, **kwargs
            )
            ax.set_title("C-factor [-]")
            return fig, ax

        self._cfactor.plot = plot

    @property
    def buffers(self):
        """Getter buffers attribute.

        For documentation, see :ref:`here <watemsedem:buffermap>`
        """
        return self._buffers

    @buffers.setter
    def buffers(self, raster):
        """Setter

        Parameters
        ----------
        raster: pathlib.Path | str
        """
        self._buffers = self.raster_factory(raster, flag_mask=False)
        # checks on raster data
        valid_non_nan(self.buffers.arr)
        valid_array_type(self.buffers.arr, required_type=np.int16)
        valid_boundaries(self.buffers.arr, lower=0, upper=None)
        valid_raster_properties_raster_with_template(self.rp, raster, epsg=self.rp.epsg)

        labels = ["No buffer", "Buffer"]

        def plot(fig=None, ax=None, *args, **kwargs):
            """Plot for buffers

            Parameters
            ----------
            fig: matplotlib.figure.Figure, default = None
                if not given, defaults to generating new figure
            ax: matplotlib.pyplot.axis, default = None
                if not given, defaults to generating new axis

            Returns
            -------
            fig: matplotlib.figure.Figure

            ax: matplotlib.axes.Axes
            """
            fig, ax = axes_creator(fig, ax)
            arr = mask_array_with_val(self.buffers.arr, self.mask.arr, 0)
            # binary plotting: if larger than 0, it is a buffer id!
            arr[arr > 0] = 1  # binary plotting!
            fig, ax = plot_discrete_raster(
                fig, ax, arr, self.rp.bounds, labels, *args, **kwargs
            )
            ax.set_title("Buffers")
            return fig, ax

        self._buffers.plot = plot

    @property
    def dtm(self):
        """Getter dtm attribute.

        For documentation, see :ref:`here <watemsedem:dtmmap>`
        """
        return self._dtm

    @dtm.setter
    def dtm(self, raster):
        """Setter

        Parameters
        ----------
        raster: pathlib.Path | str
        """
        self._dtm = self.raster_factory(raster, flag_mask=False)
        # checks on raster data
        valid_non_nan(self.dtm.arr)
        valid_array_type(self.dtm.arr, required_type=np.float32)
        # crucial that code does NOT run when nodata values detected!
        valid_nodata(self.dtm.arr, nodata_value=-9999)
        valid_nodata(self.dtm.arr, nodata_value=-99999)
        valid_boundaries(self.dtm.arr, lower=-431, upper=9000)
        # +- lowest and highest points on earth
        valid_raster_properties_raster_with_template(self.rp, raster, epsg=self.rp.epsg)

        def plot(fig=None, ax=None, *args, **kwargs):
            """Plot for DTM with cividis as colormap

            Parameters
            ----------
            fig: matplotlib.figure.Figure, default = None
                if not given, defaults to generating new figure
            ax: matplotlib.pyplot.axis, default = None
                if not given, defaults to generating new axis

            Returns
            -------
            fig: matplotlib.figure.Figure

            ax: matplotlib.axes.Axes
            """
            fig, ax = axes_creator(fig, ax)
            arr = mask_array_with_val(self.dtm.arr, self.mask.arr, 0)
            fig, ax = plot_continuous_raster(
                fig, ax, arr, self.rp.bounds, *args, **kwargs
            )
            ax.set_title("DTM [m MSL]")
            return fig, ax

        self._dtm.plot = plot

    @property
    def kfactor(self):
        """Getter kfactor attribute.
        For documentation, see :ref:`here <watemsedem:kmap>`
        """
        return self._kfactor

    @kfactor.setter
    def kfactor(self, raster_input):
        """Setter

        Parameters
        ----------
        raster: pathlib.Path | str
        """
        raster = self.raster_factory(raster_input, flag_mask=False)
        # checks on raster data
        valid_non_nan(raster.arr)
        # NO need for checking no data, deal with this in plotting!
        valid_array_type(raster.arr, required_type=np.int16)
        valid_boundaries(
            raster.arr[raster.arr != -9999], lower=0, upper=None
        )  # No data value excluded from check
        valid_raster_properties_raster_with_template(
            self.rp, raster_input, epsg=self.rp.epsg
        )
        self._kfactor = raster

        def plot(fig=None, ax=None, *args, **kwargs):
            """Plot for K-factor with cividis as colormap

            Parameters
            ----------
            fig: matplotlib.figure.Figure, default = None
                if not given, defaults to generating new figure
            ax: matplotlib.pyplot.axis, default = None
                if not given, defaults to generating new axis

            Returns
            -------
            fig: matplotlib.figure.Figure

            ax: matplotlib.axes.Axes
            """
            fig, ax = axes_creator(fig, ax)
            arr = mask_array_with_val(self.kfactor.arr, self.mask.arr, 0)
            fig, ax = plot_continuous_raster(
                fig, ax, arr, self.rp.bounds, *args, **kwargs
            )
            ax.set_title("K-factor [kg.h]/[MJ.mm]")
            return fig, ax

        self._kfactor.plot = plot

    @property
    def ktc(self):
        """Getter kTC parameter attribute.

        For documentation, see :ref:`here <watemsedem:ktcmap>`
        """
        return self._ktc

    @ktc.setter
    def ktc(self, raster):
        """Setter

        Parameters
        ----------
        raster: pathlib.Path | str
        """
        self._ktc = self.raster_factory(raster, flag_mask=False)
        # checks on raster data
        valid_non_nan(self.ktc.arr)
        valid_array_type(self.ktc.arr, required_type=np.float32)
        valid_boundaries(
            self.ktc.arr[self.ktc.arr != 9999], lower=0, upper=20
        )  # 0 to 20 if not nodata
        valid_values(self.ktc.arr[self._ktc.arr > 20], [9999])
        valid_raster_properties_raster_with_template(self.rp, raster, epsg=self.rp.epsg)

        label = ["9999"]

        def plot(fig=None, ax=None, *args, **kwargs):
            """Plot for kTC parameter with cividis as colormap

            Parameters
            ----------
            fig: matplotlib.figure.Figure, default = None
                if not given, defaults to generating new figure
            ax: matplotlib.pyplot.axis, default = None
                if not given, defaults to generating new axis

            Returns
            -------
            fig: matplotlib.figure.Figure

            ax: matplotlib.axes.Axes
            """
            fig, ax = axes_creator(fig, ax)
            arr = mask_array_with_val(self.ktc.arr, self.mask.arr, 0)
            arr_nodata = mask_array_with_val(arr, arr, 9999)
            fig, ax = plot_continuous_raster(
                fig, ax, arr_nodata, self.rp.bounds, *args, **kwargs
            )
            arr_only_nodata = np.ma.masked_where(arr != 9999, arr)
            fig, ax = plot_discrete_raster(
                fig, ax, arr_only_nodata, self.rp.bounds, label, "Set1"
            )
            ax.set_title("kTC [m]")
            return fig, ax

        self._ktc.plot = plot

    @property
    def outlet(self):
        """Getter outlet attribute.

        For documentation, see :ref:`here <watemsedem:outletmap>`
        """
        return self._outlet

    @outlet.setter
    def outlet(self, raster):
        """Setter

        Parameters
        ----------
        raster: pathlib.Path | str
        """
        self._outlet = self.raster_factory(raster, flag_mask=False)
        # checks on raster data
        valid_non_nan(self.outlet.arr)
        valid_nodata(self.outlet.arr)
        valid_array_type(self.outlet.arr, required_type=np.int16)
        valid_values(self.outlet.arr, unique_values=[0, 1])
        valid_raster_properties_raster_with_template(self.rp, raster, epsg=self.rp.epsg)

    @property
    def pfactor(self):
        """Getter P-factor attribute.

        For documentation, see :ref:`here <watemsedem:pmap>`
        """
        return self._pfactor

    @pfactor.setter
    def pfactor(self, raster):
        """Setter

        Parameters
        ----------
        raster: pathlib.Path | str
        """
        self._pfactor = self.raster_factory(raster, flag_mask=False)
        # checks on raster data
        valid_non_nan(self.pfactor.arr)
        valid_nodata(self.pfactor.arr)  # since only binary values of use!
        lower, upper = 0, 1
        valid_boundaries(self.pfactor.arr, lower=lower, upper=upper)
        valid_array_type(self.pfactor.arr, required_type=np.float32)
        valid_raster_properties_raster_with_template(self.rp, raster, epsg=self.rp.epsg)

        def plot(fig=None, ax=None, *args, **kwargs):
            """Plot for P-factor

             Parameters
            ----------
            fig: matplotlib.figure.Figure, default = None
                if not given, defaults to generating new figure
            ax: matplotlib.pyplot.axis, default = None
                if not given, defaults to generating new axis

            Returns
            -------
            fig: matplotlib.figure.Figure

            ax: matplotlib.axes.Axes
            """
            fig, ax = axes_creator(fig, ax)
            arr = mask_array_with_val(self.pfactor.arr, self.mask.arr, None)
            fig, ax = plot_continuous_raster(
                fig, ax, arr, self.rp.bounds, vmin=lower, vmax=upper, *args, **kwargs
            )
            ax.set_title("P-factor [-]")
            return fig, ax

        self._pfactor.plot = plot

    @property
    def compositelanduse(self):
        """Getter landuseparcels attribute.

        For documentation, see :ref:`here <watemsedem:prcmap>`
        """
        return self._compositelanduse

    @compositelanduse.setter
    def compositelanduse(self, raster):
        """Setter

        Parameters
        ----------
        raster: pathlib.Path | str
        """
        self._compositelanduse = self.raster_factory(raster, flag_mask=False)

        # checks
        valid_non_nan(self.compositelanduse.arr)
        valid_array_type(self.compositelanduse.arr, required_type=np.int16)
        valid_boundaries(self.compositelanduse.arr, lower=-32757, upper=32757)
        valid_raster_properties_raster_with_template(self.rp, raster, epsg=self.rp.epsg)

        # custom plotting features
        colormap = colors.ListedColormap(
            [
                "#64cf1b",
                "#3b7db4",
                "#71b651",
                "#387b00",
                "#000000",
                "#00bfff",
                "#ffffff",
                "#a47158",
            ]
        )
        labels = [
            "Grass strips (-6)",
            "Pools (-5)",
            "Meadow (-4)",
            "Forest (-3)",
            "Infrastructure (-2)",
            "River (-1)",
            "Outside boundaries (0)",
            "Agriculture (>0)",
        ]

        def plot(fig=None, ax=None, *args, **kwargs):
            """Plot for Landuseparecles

             Parameters
            ----------
            fig: matplotlib.figure.Figure, default = None
                if not given, defaults to generating new figure
            ax: matplotlib.pyplot.axis, default = None
                if not given, defaults to generating new axis

            Returns
            -------
            fig: matplotlib.figure.Figure

            ax: matplotlib.axes.Axes
            """
            fig, ax = axes_creator(fig, ax)
            arr = mask_array_with_val(self.compositelanduse.arr, self.mask.arr, 0)
            arr[arr > 0] = 1
            fig, ax = plot_discrete_raster(
                fig, ax, arr, self.rp.bounds, labels, colormap, *args, **kwargs
            )
            ax.set_title("Land use parcels")
            return fig, ax

        self._compositelanduse.plot = plot

    @property
    def ptef(self):
        """Getter ptef attribute.

        For documentation, see :ref:`here <watemsedem:parceltrapppingcrop>`
        """
        return self._ptef

    @ptef.setter
    def ptef(self, raster):
        """Setter

        Parameters
        ----------
        raster: pathlib.Path | str
        """
        self._ptef = self.raster_factory(raster, flag_mask=False)
        # checks on raster data
        valid_non_nan(self.ptef.arr)
        valid_array_type(self.ptef.arr, required_type=np.int16)
        # int16, maar waarom niet float32?
        lower, upper = 0, 100
        valid_boundaries(self.ptef.arr, lower=lower, upper=upper)
        valid_raster_properties_raster_with_template(self.rp, raster, epsg=self.rp.epsg)

        def plot(fig=None, ax=None, *args, **kwargs):
            """Plot for PTEF

            Parameters
            ----------
            fig: matplotlib.figure.Figure, default = None
                if not given, defaults to generating new figure
            ax: matplotlib.pyplot.axis, default = None
                if not given, defaults to generating new axis

            Returns
            -------
            fig: matplotlib.figure.Figure

            ax: matplotlib.axes.Axes

            """
            fig, ax = axes_creator(fig, ax)
            arr = mask_array_with_val(self.ptef.arr, self.mask.arr, 0)
            fig, ax = plot_continuous_raster(
                fig, ax, arr, self.rp.bounds, vmin=lower, vmax=upper, *args, **kwargs
            )
            ax.set_title("PTEF [%]")
            return fig, ax

        self._ptef.plot = plot

    @property
    def riversegments(self):
        """Getter riversegments attribute.

        For documentation, see :ref:`here <watemsedem:riversegmentfile>`
        """
        return self._riversegments

    @riversegments.setter
    def riversegments(self, raster):
        """Setter

        Parameters
        ----------
        raster: pathlib.Path | str
        """
        self._riversegments = self.raster_factory(raster, flag_mask=False)

        # checks
        valid_non_nan(self.ptef.arr)
        valid_array_type(self.ptef.arr, required_type=np.int16)
        valid_boundaries(self.ptef.arr, lower=0, upper=None)
        valid_raster_properties_raster_with_template(self.rp, raster, epsg=self.rp.epsg)

        def plot(fig=None, ax=None, *args, **kwargs):
            """Plot for riversegments

            Parameters
            ----------
            fig: matplotlib.figure.Figure, default = None
                if not given, defaults to generating new figure
            ax: matplotlib.pyplot.axis, default = None
                if not given, defaults to generating new axis

            Returns
            -------
            fig: matplotlib.figure.Figure

            ax: matplotlib.axes.Axes
            """
            fig, ax = axes_creator(fig, ax)
            arr = mask_array_with_val(self.riversegments.arr, self.mask.arr, 0)
            arr_only_rivers = np.ma.masked_where(arr == 0, arr)
            fig, ax = plot_continuous_raster(
                fig, ax, arr_only_rivers, self.rp.bounds, *args, **kwargs
            )
            arr_only_norivers = np.ma.masked_where(arr != 0, arr)
            fig, ax = plot_continuous_raster(
                fig,
                ax,
                arr_only_norivers,
                self.rp.bounds,
                cmap="summer",
                colorbar=False,
                alpha=0.8,
            )
            ax.set_title("Riversegments")

            return fig, ax

        self._riversegments.plot = plot

    @property
    def riverrouting(self):
        """Getter riverrouting attribute.

        For documentation, see :ref:`here <watemsedem:routingmap>`
        """
        return self._riverrouting

    @riverrouting.setter
    def riverrouting(self, raster):
        """Setter

        Parameters
        ----------
        raster: pathlib.Path | str
        """
        self._riverrouting = self.raster_factory(raster, flag_mask=False)

        # checks
        valid_non_nan(self.riverrouting.arr)
        valid_array_type(self.riverrouting.arr, required_type=np.int16)
        valid_values(
            self.riverrouting.arr[self.riverrouting.arr != -9999],
            unique_values=np.arange(0, 9).tolist(),
        )
        valid_raster_properties_raster_with_template(self.rp, raster, epsg=self.rp.epsg)
        colormap_start = mpl.cm.get_cmap("tab10")
        colorlist = []
        for i in range(0, 9):
            color = mpl.colors.rgb2hex(colormap_start(i))
            colorlist.append(color)
        colormap_routing = colors.ListedColormap(colorlist)
        labels = [
            "0 (no routing)",
            "1 (North)",
            "2 (NorthEast)",
            "3 (East)",
            "4 (SouthEast)",
            "5 (South)",
            "6 (SouthWest)",
            "7 (West)",
            "8 (NorthWest)",
        ]
        np.arange(0, 9).tolist()

        def plot(fig=None, ax=None, *args, **kwargs):
            """Plot for riverrouting

            Parameters
            ----------
            fig: matplotlib.figure.Figure, default = None
                if not given, defaults to generating new figure
            ax: matplotlib.pyplot.axis, default = None
                if not given, defaults to generating new axis

            Returns
            -------
            fig: matplotlib.figure.Figure

            ax: matplotlib.axes.Axes
            """
            fig, ax = axes_creator(fig, ax)
            arr = mask_array_with_val(self.riverrouting.arr, self.mask.arr, 0)
            arr[arr == -9999] = 0  # no routing
            fig, ax = plot_discrete_raster(
                fig, ax, arr, self.rp.bounds, labels, colormap_routing, *args, **kwargs
            )
            ax.set_title("Riverrouting")
            return fig, ax

        self._riverrouting.plot = plot

    @property
    def sewers(self):
        """Getter sewers attribute.

        For documentation, see :ref:`here <watemsedem:sewermapfile>`
        """
        return self._sewers

    @sewers.setter
    def sewers(self, raster):
        """Setter

        Parameters
        ----------
        raster: pathlib.Path | str
        """
        self._sewers = self.raster_factory(raster, flag_mask=False)

        # checks
        valid_non_nan(self.sewers.arr)
        valid_array_type(self.sewers.arr, required_type=np.float32)
        lower, upper = 0, 1
        valid_boundaries(self.sewers.arr, lower=lower, upper=upper)
        valid_raster_properties_raster_with_template(self.rp, raster, epsg=self.rp.epsg)

        def plot(fig=None, ax=None, *args, **kwargs):
            """Plot for sewers

            Parameters
            ----------
            fig: matplotlib.figure.Figure, default = None
                if not given, defaults to generating new figure
            ax: matplotlib.pyplot.axis, default = None
                if not given, defaults to generating new axis

            Returns
            -------
            fig: matplotlib.figure.Figure

            ax: matplotlib.axes.Axes
            """
            fig, ax = axes_creator(fig, ax)
            arr = mask_array_with_val(self.sewers.arr, self.mask.arr, 0)
            fig, ax = plot_continuous_raster(
                fig, ax, arr, self.rp.bounds, vmin=lower, vmax=upper, *args, **kwargs
            )
            ax.set_title("Fraction of sediment in sewers")

            return fig, ax

        self._sewers.plot = plot

    @property
    def upstreamsegments(self):
        """Getter upstreamsegments dataframe.

        For documentation, see :ref:`here <watemsedem:upstrsegments>`
        """
        return self._upstreamsegments

    @upstreamsegments.setter
    @valid_segments
    def upstreamsegments(self, text):
        """Setter

        Parameters
        ----------
        text: pathlib.Path | str
        """
        self._upstreamsegments = pd.read_table(text)
        # checks
        array = self.upstreamsegments[["line_id", "upstream_line"]].values
        valid_non_nan(array)
        valid_values(array, unique_values=np.unique(self.riversegments.arr).tolist())
        valid_array_type(array, required_type=np.int64)

    @property
    def adjacentsegments(self):
        """Getter adjacentsegments dataframe.

        For documentation, see :ref:`here <watemsedem:adjsegments>`
        """
        return self._adjacentsegments

    @adjacentsegments.setter
    @valid_segments
    def adjacentsegments(self, text):
        """Setter

        Parameters
        ----------
        text: pathlib.Path | str
        """
        self._adjacentsegments = pd.read_table(text)
        # checks
        array = self.adjacentsegments.values
        valid_non_nan(array)
        valid_values(array, unique_values=np.unique(self.riversegments.arr).tolist())
        valid_array_type(array, required_type=np.int64)



import numpy as np
import pytest

from pywatemsedem.io.valid import (
    valid_array_type,
    valid_boundaries,
    valid_nodata,
    valid_non_nan,
    valid_values,
)


def test_check_boundaries():
    """test for check_boundaries"""
    arr = np.array([[0, 1, 2, np.nan], [0, 1, 3, 4]])

    # proper numeric bounds
    lower = 0
    upper = 4
    assert valid_boundaries(arr, lower, upper)

    # invalid nan bounds: allowable
    lower = np.nan
    upper = np.nan
    assert valid_boundaries(arr, lower, upper)

    # invalid type inputted for lower and upper
    lower = 0
    upper = "4"
    with pytest.raises(TypeError) as excinfo:
        valid_boundaries(arr, lower, upper)
    assert "Upper boundary should be numeric." in str(excinfo.value)
    lower = "0"
    upper = 4
    with pytest.raises(TypeError) as excinfo:
        valid_boundaries(arr, lower, upper)
    assert "Lower boundary should be numeric." in str(excinfo.value)

    # None bounds
    lower = None
    upper = None
    with pytest.warns(UserWarning, match="No lower or upper boundaries are"):
        valid_boundaries(arr, lower, upper)

    # wrong lower/upper bound
    upper = 2
    with pytest.raises(ValueError) as excinfo:
        valid_boundaries(arr, upper=upper)
    assert "are higher than upper bound" in str(excinfo.value)
    lower = 2
    with pytest.raises(ValueError) as excinfo:
        valid_boundaries(arr, lower=lower)
    assert "are lower than lower bound" in str(excinfo.value)


def test_check_unique_values():
    """test for check_unique_values"""
    arr = np.array([[0, 1, 2, 0], [0, 1, 3, 4]])
    unique_values = [0, 1, 2, 3, 4, 5]
    assert valid_values(arr, unique_values)

    # unique values not known in array
    arr2 = np.concatenate((arr, np.array([5, 6, 7, 8]).reshape(-1, 4)))
    with pytest.raises(ValueError) as excinfo:
        valid_values(arr2, unique_values)
    assert " values that are not present in unique_values" in str(excinfo.value)


def test_check_nan():
    """test for check_nan"""
    arr = np.array([[0, 1, 2, 0], [0, 1, 3, 4]], dtype=np.float64)
    assert valid_non_nan(arr)
    # input nan
    arr[1, 3] = np.nan
    with pytest.raises(ValueError) as excinfo:
        valid_non_nan(arr)
    assert "can not contain nan values" in str(excinfo.value)


def test_valid_array_type():
    """test for check_datatype"""
    arr = np.array([[0, 1, 2, 0], [0, 1, 3, 4]], dtype=np.int32)
    assert valid_array_type(arr, np.int32)

    # wrong datatype required
    with pytest.raises(ValueError) as excinfo:
        valid_array_type(arr, np.float32)
    assert "is not required type" in str(excinfo.value)

    # wrong datatype array
    arr_str = arr.astype(str)
    with pytest.raises(ValueError) as excinfo:
        valid_array_type(arr_str, np.int32)
    assert "is not required type" in str(excinfo.value)

    # string instead of np.dtype: not allowed
    req_type = "string"
    with pytest.raises(TypeError) as excinfo:
        valid_array_type(arr, req_type)
    assert "Use numpy dtype as required type." in str(excinfo.value)


def test_valid_nodata():
    """test for check_nodata"""
    arr = np.array([[0, 1, 2, 0], [0, 1, 3, 4]], dtype=np.float32)
    valid_nodata(arr, nodata_value=-9999)
    arr[1, 2] = -9999
    with pytest.raises(Warning) as excinfo:
        valid_nodata(arr)
    assert "A nodata value" in str(excinfo.value)
    # nan as nodatavalue
    nodata_value = np.nan
    valid_nodata(arr, nodata_value=nodata_value)
